"""Handler for rescanning a specific commit."""

from typing import TYPE_CHECKING, Any

from kodit.application.services.queue_service import QueueService
from kodit.application.services.reporting import ProgressTracker
from kodit.domain.protocols import (
    EnrichmentAssociationRepository,
    EnrichmentV2Repository,
    GitFileRepository,
)
from kodit.domain.services.bm25_service import BM25DomainService
from kodit.domain.value_objects import (
    DeleteRequest,
    PrescribedOperations,
    QueuePriority,
    TaskOperation,
    TrackableType,
)
from kodit.infrastructure.sqlalchemy import entities as db_entities
from kodit.infrastructure.sqlalchemy.embedding_repository import (
    SqlAlchemyEmbeddingRepository,
)
from kodit.infrastructure.sqlalchemy.query import FilterOperator, QueryBuilder

if TYPE_CHECKING:
    from kodit.application.services.enrichment_query_service import (
        EnrichmentQueryService,
    )


class RescanCommitHandler:
    """Handler for rescanning a specific commit."""

    def __init__(  # noqa: PLR0913
        self,
        git_file_repository: GitFileRepository,
        bm25_service: BM25DomainService,
        embedding_repository: SqlAlchemyEmbeddingRepository,
        enrichment_v2_repository: EnrichmentV2Repository,
        enrichment_association_repository: EnrichmentAssociationRepository,
        enrichment_query_service: "EnrichmentQueryService",
        queue: QueueService,
        operation: ProgressTracker,
    ) -> None:
        """Initialize the rescan commit handler."""
        self.git_file_repository = git_file_repository
        self.bm25_service = bm25_service
        self.embedding_repository = embedding_repository
        self.enrichment_v2_repository = enrichment_v2_repository
        self.enrichment_association_repository = enrichment_association_repository
        self.enrichment_query_service = enrichment_query_service
        self.queue = queue
        self.operation = operation

    async def execute(self, payload: dict[str, Any]) -> None:
        """Execute rescan commit operation."""
        repository_id = payload["repository_id"]
        commit_sha = payload["commit_sha"]

        async with self.operation.create_child(
            TaskOperation.RESCAN_COMMIT,
            trackable_type=TrackableType.KODIT_COMMIT,
            trackable_id=repository_id,
        ):
            # Delete snippet enrichments and their indices
            await self._delete_snippet_enrichments(commit_sha)

            # Delete example enrichments and their indices
            await self._delete_example_enrichments(commit_sha)

            # Delete commit-level enrichments
            await self._delete_commit_enrichments(commit_sha)

            # Delete git files for this commit
            await self.git_file_repository.delete_by_commit_sha(commit_sha)

            # Re-trigger the indexing pipeline
            await self.queue.enqueue_tasks(
                tasks=PrescribedOperations.SCAN_AND_INDEX_COMMIT,
                base_priority=QueuePriority.USER_INITIATED,
                payload={
                    "repository_id": repository_id,
                    "commit_sha": commit_sha,
                },
            )

    async def _delete_snippet_enrichments(self, commit_sha: str) -> None:
        """Delete snippet enrichments and their indices for a commit."""
        snippet_enrichments = (
            await self.enrichment_query_service.get_all_snippets_for_commit(commit_sha)
        )
        enrichment_ids = [
            enrichment.id for enrichment in snippet_enrichments if enrichment.id
        ]

        if not enrichment_ids:
            return

        # Delete from BM25 and embedding indices
        snippet_id_strings = [str(sid) for sid in enrichment_ids]
        delete_request = DeleteRequest(snippet_ids=snippet_id_strings)
        await self.bm25_service.delete_documents(delete_request)

        for snippet_id in enrichment_ids:
            await self.embedding_repository.delete_embeddings_by_snippet_id(
                str(snippet_id)
            )

        # Delete enrichment associations for snippets
        await self.enrichment_association_repository.delete_by_query(
            QueryBuilder()
            .filter("entity_type", FilterOperator.EQ, "snippet_v2")
            .filter("entity_id", FilterOperator.IN, snippet_id_strings)
        )

        # Delete the enrichments themselves
        await self.enrichment_v2_repository.delete_by_query(
            QueryBuilder().filter("id", FilterOperator.IN, enrichment_ids)
        )

    async def _delete_example_enrichments(self, commit_sha: str) -> None:
        """Delete example enrichments and their indices for a commit."""
        example_enrichments = (
            await self.enrichment_query_service.get_all_examples_for_commit(commit_sha)
        )
        enrichment_ids = [
            enrichment.id for enrichment in example_enrichments if enrichment.id
        ]

        if not enrichment_ids:
            return

        # Delete from BM25 and embedding indices
        example_id_strings = [str(eid) for eid in enrichment_ids]
        delete_request = DeleteRequest(snippet_ids=example_id_strings)
        await self.bm25_service.delete_documents(delete_request)

        for example_id in enrichment_ids:
            await self.embedding_repository.delete_embeddings_by_snippet_id(
                str(example_id)
            )

        # Delete enrichment associations for examples
        await self.enrichment_association_repository.delete_by_query(
            QueryBuilder()
            .filter("entity_type", FilterOperator.EQ, "example_v2")
            .filter("entity_id", FilterOperator.IN, example_id_strings)
        )

        # Delete the enrichments themselves
        await self.enrichment_v2_repository.delete_by_query(
            QueryBuilder().filter("id", FilterOperator.IN, enrichment_ids)
        )

    async def _delete_commit_enrichments(self, commit_sha: str) -> None:
        """Delete commit-level enrichments for a commit."""
        existing_enrichment_associations = (
            await self.enrichment_association_repository.find(
                QueryBuilder()
                .filter(
                    "entity_type",
                    FilterOperator.EQ,
                    db_entities.GitCommit.__tablename__,
                )
                .filter("entity_id", FilterOperator.EQ, commit_sha)
            )
        )
        enrichment_ids = [a.enrichment_id for a in existing_enrichment_associations]
        if not enrichment_ids:
            return

        # Delete associations first
        await self.enrichment_association_repository.delete_by_query(
            QueryBuilder().filter("enrichment_id", FilterOperator.IN, enrichment_ids)
        )
        # Then delete enrichments
        await self.enrichment_v2_repository.delete_by_query(
            QueryBuilder().filter("id", FilterOperator.IN, enrichment_ids)
        )



# ruff: noqa
"""add generic enrichment type

Revision ID: 19f8c7faf8b9
Revises: f9e5ef5e688f
Create Date: 2025-09-29 21:38:19.093821

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa

from kodit.domain.enrichments.development.development import ENRICHMENT_TYPE_DEVELOPMENT
from kodit.domain.enrichments.development.snippet.snippet import (
    ENRICHMENT_SUBTYPE_SNIPPET_SUMMARY,
)


# revision identifiers, used by Alembic.
revision: str = "19f8c7faf8b9"
down_revision: Union[str, None] = "f9e5ef5e688f"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    """Upgrade schema."""
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "enrichments_v2",
        sa.Column("content", sa.UnicodeText(), nullable=False),
        sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column("type", sa.String(length=255), nullable=False),
        sa.Column("subtype", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            nullable=False,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            nullable=False,
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        "idx_type_subtype", "enrichments_v2", ["type", "subtype"], unique=False
    )
    op.create_index(
        op.f("ix_enrichments_v2_type"), "enrichments_v2", ["type"], unique=False
    )
    op.create_index(
        op.f("ix_enrichments_v2_subtype"), "enrichments_v2", ["subtype"], unique=False
    )
    op.create_table(
        "enrichment_associations",
        sa.Column("enrichment_id", sa.Integer(), nullable=False),
        sa.Column("entity_type", sa.String(length=50), nullable=False),
        sa.Column("entity_id", sa.String(length=255), nullable=False),
        sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            nullable=False,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["enrichment_id"], ["enrichments_v2.id"], ondelete="CASCADE"
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint(
            "entity_type", "entity_id", "enrichment_id", name="uix_entity_enrichment"
        ),
        sqlite_autoincrement=True,
    )
    op.create_index(
        "idx_entity_lookup",
        "enrichment_associations",
        ["entity_type", "entity_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_enrichment_associations_enrichment_id"),
        "enrichment_associations",
        ["enrichment_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_enrichment_associations_entity_id"),
        "enrichment_associations",
        ["entity_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_enrichment_associations_entity_type"),
        "enrichment_associations",
        ["entity_type"],
        unique=False,
    )

    # Migrate existing enrichments from old table to new structure
    # All old enrichments are snippet enrichments
    connection = op.get_bind()

    # Check if old enrichments table exists
    inspector = sa.inspect(connection)
    if "enrichments" in inspector.get_table_names():
        # Copy enrichments to new table
        old_enrichments = connection.execute(
            sa.text(
                "SELECT id, content, created_at, updated_at, snippet_sha FROM enrichments"
            )
        ).fetchall()

        for old_enrichment in old_enrichments:
            # Check if database supports RETURNING (PostgreSQL) or use lastrowid (SQLite)
            dialect_name = connection.dialect.name

            if dialect_name == "postgresql":
                result = connection.execute(
                    sa.text(
                        "INSERT INTO enrichments_v2 (type, subtype, content, created_at, updated_at) "
                        "VALUES (:type, :subtype, :content, :created_at, :updated_at) "
                        "RETURNING id"
                    ),
                    {
                        "type": ENRICHMENT_TYPE_DEVELOPMENT,
                        "subtype": ENRICHMENT_SUBTYPE_SNIPPET_SUMMARY,
                        "content": old_enrichment[1],
                        "created_at": old_enrichment[2],
                        "updated_at": old_enrichment[3],
                    },
                )
                row = result.fetchone()
                if row is None:
                    raise RuntimeError("Failed to insert enrichment")
                new_enrichment_id = row[0]
            else:
                # SQLite and other databases
                result = connection.execute(
                    sa.text(
                        "INSERT INTO enrichments_v2 (type, subtype, content, created_at, updated_at) "
                        "VALUES (:type, :subtype, :content, :created_at, :updated_at)"
                    ),
                    {
                        "type": ENRICHMENT_TYPE_DEVELOPMENT,
                        "subtype": ENRICHMENT_SUBTYPE_SNIPPET_SUMMARY,
                        "content": old_enrichment[1],
                        "created_at": old_enrichment[2],
                        "updated_at": old_enrichment[3],
                    },
                )
                new_enrichment_id = result.lastrowid

            # Insert association (snippet_v2 is the entity type for snippets)
            connection.execute(
                sa.text(
                    "INSERT INTO enrichment_associations "
                    "(enrichment_id, entity_type, entity_id, created_at, updated_at) "
                    "VALUES (:enrichment_id, :entity_type, :entity_id, :created_at, :updated_at)"
                ),
                {
                    "enrichment_id": new_enrichment_id,
                    "entity_type": "snippet_v2",
                    "entity_id": old_enrichment[4],  # snippet_sha
                    "created_at": old_enrichment[2],
                    "updated_at": old_enrichment[3],
                },
            )

        # Drop old enrichments table
        op.drop_table("enrichments")

    # ### end Alembic commands ###


def downgrade() -> None:
    """Downgrade schema."""
    # ### commands auto generated by Alembic - please adjust! ###
    # Recreate old enrichments table
    op.create_table(
        "enrichments",
        sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            nullable=False,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            nullable=False,
        ),
        sa.Column("snippet_sha", sa.String(length=64), nullable=False),
        sa.Column("type", sa.String(length=50), nullable=False),
        sa.Column("content", sa.UnicodeText(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
        sa.ForeignKeyConstraint(["snippet_sha"], ["snippets_v2.sha"]),
    )
    op.create_index(
        op.f("ix_enrichments_snippet_sha"), "enrichments", ["snippet_sha"], unique=False
    )
    op.create_index(op.f("ix_enrichments_type"), "enrichments", ["type"], unique=False)

    # Migrate data back from new structure to old table
    connection = op.get_bind()

    # Get enrichments with their associations
    enrichments_with_associations = connection.execute(
        sa.text(
            "SELECT e.id, e.content, e.created_at, e.updated_at, "
            "a.entity_id, a.entity_type "
            "FROM enrichments_v2 e "
            "JOIN enrichment_associations a ON e.id = a.enrichment_id "
            "WHERE a.entity_type = 'snippet_v2'"
        )
    ).fetchall()

    for enrichment in enrichments_with_associations:
        connection.execute(
            sa.text(
                "INSERT INTO enrichments "
                "(content, created_at, updated_at, snippet_sha, type) "
                "VALUES (:content, :created_at, :updated_at, :snippet_sha, :type)"
            ),
            {
                "content": enrichment[1],
                "created_at": enrichment[2],
                "updated_at": enrichment[3],
                "snippet_sha": enrichment[4],
                "type": "summarization",  # Default to summarization type
            },
        )

    op.drop_index(
        op.f("ix_enrichment_associations_entity_type"),
        table_name="enrichment_associations",
    )
    op.drop_index(
        op.f("ix_enrichment_associations_entity_id"),
        table_name="enrichment_associations",
    )
    op.drop_index(
        op.f("ix_enrichment_associations_enrichment_id"),
        table_name="enrichment_associations",
    )
    op.drop_index("idx_entity_lookup", table_name="enrichment_associations")
    op.drop_table("enrichment_associations")
    op.drop_index(op.f("ix_enrichments_v2_subtype"), table_name="enrichments_v2")
    op.drop_index(op.f("ix_enrichments_v2_type"), table_name="enrichments_v2")
    op.drop_index("idx_type_subtype", table_name="enrichments_v2")
    op.drop_table("enrichments_v2")
    # ### end Alembic commands ###



"""Test for GitPython zombie process leak on clone failure (Linux only)."""

import contextlib
import subprocess
import tempfile
from pathlib import Path

import pytest

from kodit.infrastructure.cloning.git.git_python_adaptor import GitPythonAdapter

FAILING_URL = "https://api:invalid@app.helix.ml/git/nonexistent"


def count_git_zombies() -> int:
    """Count zombie git processes via ps aux."""
    result = subprocess.run(  # noqa: S603
        ["/bin/ps", "aux"], capture_output=True, text=True, check=False
    )
    return sum(
        1
        for line in result.stdout.split("\n")
        if "defunct" in line and "git" in line.lower()
    )


@pytest.mark.asyncio
async def test_failed_clone_does_not_leave_zombie_processes() -> None:
    """Cloning a repo that fails should not leave zombie git processes."""
    adapter = GitPythonAdapter(max_workers=1)
    baseline = count_git_zombies()

    with tempfile.TemporaryDirectory() as tmpdir:
        for i in range(3):
            with contextlib.suppress(Exception):
                await adapter.clone_repository(FAILING_URL, Path(tmpdir) / f"clone-{i}")

    new_zombies = count_git_zombies() - baseline
    assert new_zombies == 0, (
        f"Clone failures created {new_zombies} zombie git process(es)"
    )


@pytest.mark.asyncio
async def test_dulwich_failed_clone_does_not_leave_zombie_processes() -> None:
    """Cloning a repo that fails with Dulwich should not leave zombie processes."""
    from kodit.infrastructure.cloning.git.dulwich_adaptor import DulwichAdapter

    adapter = DulwichAdapter(max_workers=1)
    baseline = count_git_zombies()

    with tempfile.TemporaryDirectory() as tmpdir:
        for i in range(3):
            with contextlib.suppress(Exception):
                await adapter.clone_repository(FAILING_URL, Path(tmpdir) / f"clone-{i}")

    new_zombies = count_git_zombies() - baseline
    assert new_zombies == 0, (
        f"Dulwich clone failures created {new_zombies} zombie git process(es)"
    )


@pytest.mark.asyncio
async def test_pygit2_failed_clone_does_not_leave_zombie_processes() -> None:
    """Cloning a repo that fails with PyGit2 should not leave zombie processes."""
    from kodit.infrastructure.cloning.git.pygit2_adaptor import PyGit2Adapter

    adapter = PyGit2Adapter(max_workers=1)
    baseline = count_git_zombies()

    with tempfile.TemporaryDirectory() as tmpdir:
        for i in range(3):
            with contextlib.suppress(Exception):
                await adapter.clone_repository(FAILING_URL, Path(tmpdir) / f"clone-{i}")

    new_zombies = count_git_zombies() - baseline
    assert new_zombies == 0, (
        f"PyGit2 clone failures created {new_zombies} zombie git process(es)"
    )



"""Base query for SQLAlchemy repositories."""

from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum
from typing import Any, Self

from sqlalchemy import Select

from kodit.domain.enrichments.enrichment import EnrichmentV2
from kodit.infrastructure.api.v1.query_params import PaginationParams
from kodit.infrastructure.sqlalchemy import entities as db_entities


class Query(ABC):
    """Base query/specification object for encapsulating query logic."""

    @abstractmethod
    def apply(self, stmt: Select, model_type: type) -> Select:
        """Apply this query's criteria to a SQLAlchemy Select statement."""


class FilterOperator(Enum):
    """SQL filter operators."""

    EQ = "eq"
    NE = "ne"
    GT = "gt"
    GTE = "ge"
    LT = "lt"
    LTE = "le"
    IN = "in_"
    LIKE = "like"
    ILIKE = "ilike"


@dataclass
class FilterCriteria:
    """Filter criteria for a query."""

    field: str
    operator: FilterOperator
    value: Any

    def apply(self, model_type: type, stmt: Select) -> Select:  # noqa: C901
        """Apply filter to statement."""
        column = getattr(model_type, self.field)

        # Convert AnyUrl to string for SQLAlchemy comparison
        value = self.value
        if hasattr(value, "__str__") and type(value).__module__ == "pydantic.networks":
            value = str(value)

        # Use column comparison methods instead of operators module
        condition = None
        match self.operator:
            case FilterOperator.EQ:
                condition = column == value
            case FilterOperator.NE:
                condition = column != value
            case FilterOperator.GT:
                condition = column > value
            case FilterOperator.GTE:
                condition = column >= value
            case FilterOperator.LT:
                condition = column < value
            case FilterOperator.LTE:
                condition = column <= value
            case FilterOperator.IN:
                condition = column.in_(value)
            case FilterOperator.LIKE:
                condition = column.like(value)
            case FilterOperator.ILIKE:
                condition = column.ilike(value)

        return stmt.where(condition)


@dataclass
class SortCriteria:
    """Sort criteria for a query."""

    field: str
    descending: bool = False

    def apply(self, model_type: type, stmt: Select) -> Select:
        """Apply sort to statement."""
        column = getattr(model_type, self.field)
        return stmt.order_by(column.desc() if self.descending else column.asc())


@dataclass
class PaginationCriteria:
    """Pagination criteria for a query."""

    limit: int | None = None
    offset: int = 0

    def apply(self, stmt: Select) -> Select:
        """Apply pagination to statement."""
        stmt = stmt.offset(self.offset)
        if self.limit is not None:
            stmt = stmt.limit(self.limit)
        return stmt


class QueryBuilder(Query):
    """Composable query builder for constructing database queries."""

    DEFAULT_SORT_FIELD = "created_at"
    DEFAULT_SORT_DESCENDING = True

    def __init__(self) -> None:
        """Initialize query builder."""
        self._filters: list[FilterCriteria] = []
        self._sorts: list[SortCriteria] = []
        self._pagination: PaginationCriteria | None = None

    def filter(self, field: str, operator: FilterOperator, value: Any) -> Self:
        """Add a filter criterion."""
        self._filters.append(FilterCriteria(field, operator, value))
        return self

    def sort(self, field: str, *, descending: bool = False) -> Self:
        """Add a sort criterion."""
        self._sorts.append(SortCriteria(field, descending))
        return self

    def paginate(self, pagination: PaginationParams) -> Self:
        """Add pagination."""
        self._pagination = PaginationCriteria(
            limit=pagination.limit, offset=pagination.offset
        )
        return self

    def apply_filters_only(self, stmt: Select, model_type: type) -> Select:
        """Apply only filter criteria to the statement."""
        for filter_criteria in self._filters:
            stmt = filter_criteria.apply(model_type, stmt)
        return stmt

    def apply(self, stmt: Select, model_type: type) -> Select:
        """Apply all criteria to the statement."""
        for filter_criteria in self._filters:
            stmt = filter_criteria.apply(model_type, stmt)

        if not self._sorts:
            self._sorts = [
                SortCriteria(
                    field=self.DEFAULT_SORT_FIELD,
                    descending=self.DEFAULT_SORT_DESCENDING,
                )
            ]
        for sort_criteria in self._sorts:
            stmt = sort_criteria.apply(model_type, stmt)

        if self._pagination:
            stmt = self._pagination.apply(stmt)

        return stmt

    def get_large_in_filters(self, chunk_size: int) -> list[FilterCriteria]:
        """Get filters that have IN clauses larger than chunk_size."""
        return [
            f
            for f in self._filters
            if f.operator == FilterOperator.IN
            and isinstance(f.value, list)
            and len(f.value) > chunk_size
        ]

    def with_replaced_filter(
        self, old_filter: FilterCriteria, new_filter: FilterCriteria
    ) -> Self:
        """Create a copy of this query with a filter replaced."""
        new_query = self.__class__()
        # Copy internal state - accessing private members is intentional for cloning
        new_query._filters = [  # noqa: SLF001
            new_filter if f == old_filter else f for f in self._filters
        ]
        new_query._sorts = self._sorts.copy()  # noqa: SLF001
        new_query._pagination = self._pagination  # noqa: SLF001
        return new_query  # type: ignore[return-value]


class EnrichmentAssociationQueryBuilder(QueryBuilder):
    """Query builder for enrichment association entities."""

    def for_enrichment_ids(self, enrichment_ids: list[int]) -> Self:
        """Build a query for associations by enrichment IDs."""
        self.filter(
            db_entities.EnrichmentAssociation.enrichment_id.key,
            FilterOperator.IN,
            enrichment_ids,
        )
        return self

    def for_enrichments(self, enrichments: list[EnrichmentV2]) -> Self:
        """Build a query for enrichment associations by entity IDs."""
        self.filter(
            db_entities.EnrichmentAssociation.enrichment_id.key,
            FilterOperator.IN,
            [enrichment.id for enrichment in enrichments if enrichment.id is not None],
        )
        return self

    def for_enrichment_type(self) -> Self:
        """Build a query for enrichment types."""
        return self.for_entity_type(db_entities.EnrichmentV2.__tablename__)

    def for_entity_type(self, entity_type: str) -> Self:
        """Build a query for enrichment associations by entity type."""
        self.filter(
            db_entities.EnrichmentAssociation.entity_type.key,
            FilterOperator.EQ,
            entity_type,
        )
        return self

    def for_entity_ids(self, entity_ids: list[str]) -> Self:
        """Build a query for enrichment associations by entity IDs."""
        self.filter(
            db_entities.EnrichmentAssociation.entity_id.key,
            FilterOperator.IN,
            entity_ids,
        )
        return self

    @staticmethod
    def for_enrichment_association(
        entity_type: str,
        entity_id: str,
    ) -> QueryBuilder:
        """Build a query for a specific enrichment association."""
        return EnrichmentAssociationQueryBuilder.for_enrichment_associations(
            entity_type,
            [entity_id],
        )

    @staticmethod
    def for_enrichment_associations(
        entity_type: str, entity_ids: list[str]
    ) -> QueryBuilder:
        """Build a query for enrichment associations by entity type and IDs."""
        return (
            QueryBuilder()
            .filter(
                db_entities.EnrichmentAssociation.entity_type.key,
                FilterOperator.EQ,
                entity_type,
            )
            .filter(
                db_entities.EnrichmentAssociation.entity_id.key,
                FilterOperator.IN,
                entity_ids,
            )
        )

    @staticmethod
    def type_and_ids(
        entity_type: str,
        enrichment_ids: list[int],
    ) -> QueryBuilder:
        """Build a query for enrichment associations by enrichment IDs."""
        return (
            QueryBuilder()
            .filter(
                db_entities.EnrichmentAssociation.entity_type.key,
                FilterOperator.EQ,
                entity_type,
            )
            .filter(
                db_entities.EnrichmentAssociation.enrichment_id.key,
                FilterOperator.IN,
                enrichment_ids,
            )
        )

    @staticmethod
    def associations_pointing_to_these_enrichments(
        enrichment_ids: list[int],
    ) -> QueryBuilder:
        """Build a query for enrichment associations pointing to these enrichments."""
        return EnrichmentAssociationQueryBuilder.type_and_ids(
            entity_type=db_entities.EnrichmentV2.__tablename__,
            enrichment_ids=enrichment_ids,
        )

    def for_commit(self, commit_sha: str) -> Self:
        """Build a query for enrichment associations for a commit."""
        self.filter(
            db_entities.EnrichmentAssociation.entity_type.key,
            FilterOperator.EQ,
            db_entities.GitCommit.__tablename__,
        )
        self.filter(
            db_entities.EnrichmentAssociation.entity_id.key,
            FilterOperator.EQ,
            commit_sha,
        )
        return self


class EnrichmentQueryBuilder(QueryBuilder):
    """Query builder for enrichment entities."""

    def for_ids(self, enrichment_ids: list[int]) -> Self:
        """Build a query for enrichments by their IDs."""
        self.filter(
            db_entities.EnrichmentV2.id.key,
            FilterOperator.IN,
            enrichment_ids,
        )
        return self

    def for_type(self, enrichment_type: str) -> Self:
        """Build a query for enrichments by their type."""
        self.filter(
            db_entities.EnrichmentV2.type.key,
            FilterOperator.EQ,
            enrichment_type,
        )
        return self

    def for_subtype(self, enrichment_subtype: str) -> Self:
        """Build a query for enrichments by their subtype."""
        self.filter(
            db_entities.EnrichmentV2.subtype.key,
            FilterOperator.EQ,
            enrichment_subtype,
        )
        return self


class GitFileQueryBuilder(QueryBuilder):
    """Query builder for git file entities."""

    def for_commit_sha(self, commit_sha: str) -> Self:
        """Build a query for git files by their commit SHA."""
        self.filter(
            db_entities.GitCommitFile.commit_sha.key,
            FilterOperator.EQ,
            commit_sha,
        )
        return self

    def for_blob_sha(self, blob_sha: str) -> Self:
        """Build a query for git files by their blob SHA."""
        self.filter(
            db_entities.GitCommitFile.blob_sha.key,
            FilterOperator.EQ,
            blob_sha,
        )
        return self


class TaskStatusQueryBuilder(QueryBuilder):
    """Query builder for task status entities."""

    def for_repository(self, repository_id: int) -> Self:
        """Build a query for task statuses by repository."""
        from kodit.domain.value_objects import TrackableType

        self.filter(
            db_entities.TaskStatus.trackable_type.key,
            FilterOperator.EQ,
            str(TrackableType.KODIT_REPOSITORY),
        )
        self.filter(
            db_entities.TaskStatus.trackable_id.key,
            FilterOperator.EQ,
            repository_id,
        )
        return self



"""Create a big object that contains all the application services."""

from collections.abc import Callable

from sqlalchemy.ext.asyncio import AsyncSession

from kodit.application.factories.reporting_factory import create_server_operation
from kodit.application.handlers.commit.api_docs import ApiDocsHandler
from kodit.application.handlers.commit.architecture_discovery import (
    ArchitectureDiscoveryHandler,
)
from kodit.application.handlers.commit.commit_description import (
    CommitDescriptionHandler,
)
from kodit.application.handlers.commit.cookbook import CookbookHandler
from kodit.application.handlers.commit.create_bm25_index import (
    CreateBM25IndexHandler,
)
from kodit.application.handlers.commit.create_code_embeddings import (
    CreateCodeEmbeddingsHandler,
)
from kodit.application.handlers.commit.create_example_code_embeddings import (
    CreateExampleCodeEmbeddingsHandler,
)
from kodit.application.handlers.commit.create_example_summary import (
    CreateExampleSummaryHandler,
)
from kodit.application.handlers.commit.create_example_summary_embeddings import (
    CreateExampleSummaryEmbeddingsHandler,
)
from kodit.application.handlers.commit.create_summary_embeddings import (
    CreateSummaryEmbeddingsHandler,
)
from kodit.application.handlers.commit.create_summary_enrichment import (
    CreateSummaryEnrichmentHandler,
)
from kodit.application.handlers.commit.database_schema import DatabaseSchemaHandler
from kodit.application.handlers.commit.extract_examples import ExtractExamplesHandler
from kodit.application.handlers.commit.extract_snippets import ExtractSnippetsHandler
from kodit.application.handlers.commit.rescan_commit import RescanCommitHandler
from kodit.application.handlers.commit.scan_commit import ScanCommitHandler
from kodit.application.handlers.registry import TaskHandlerRegistry
from kodit.application.handlers.repository.clone_repository import (
    CloneRepositoryHandler,
)
from kodit.application.handlers.repository.delete_repository import (
    DeleteRepositoryHandler,
)
from kodit.application.handlers.repository.sync_repository import (
    SyncRepositoryHandler,
)
from kodit.application.services.code_search_application_service import (
    CodeSearchApplicationService,
)
from kodit.application.services.commit_indexing_application_service import (
    CommitIndexingApplicationService,
)
from kodit.application.services.enrichment_query_service import (
    EnrichmentQueryService,
)
from kodit.application.services.queue_service import QueueService
from kodit.application.services.reporting import ProgressTracker
from kodit.application.services.repository_query_service import (
    RepositoryQueryService,
)
from kodit.application.services.repository_sync_service import RepositorySyncService
from kodit.application.services.sync_scheduler import SyncSchedulerService
from kodit.config import AppContext
from kodit.domain.enrichments.architecture.physical.formatter import (
    PhysicalArchitectureFormatter,
)
from kodit.domain.enrichments.enricher import Enricher
from kodit.domain.protocols import (
    EnrichmentAssociationRepository,
    EnrichmentV2Repository,
    FusionService,
    GitAdapter,
    GitBranchRepository,
    GitCommitRepository,
    GitFileRepository,
    GitRepoRepository,
    GitTagRepository,
    SnippetRepositoryV2,
    TaskStatusRepository,
)
from kodit.domain.services.bm25_service import BM25DomainService, BM25Repository
from kodit.domain.services.cookbook_context_service import CookbookContextService
from kodit.domain.services.embedding_service import EmbeddingDomainService
from kodit.domain.services.git_repository_service import (
    GitRepositoryScanner,
    RepositoryCloner,
)
from kodit.domain.services.physical_architecture_service import (
    PhysicalArchitectureService,
)
from kodit.domain.tracking.resolution_service import TrackableResolutionService
from kodit.infrastructure.bm25.local_bm25_repository import LocalBM25Repository
from kodit.infrastructure.bm25.vectorchord_bm25_repository import (
    VectorChordBM25Repository,
)
from kodit.infrastructure.cloning.git.factory import create_git_adapter
from kodit.infrastructure.database_schema.database_schema_detector import (
    DatabaseSchemaDetector,
)
from kodit.infrastructure.embedding.embedding_factory import (
    embedding_domain_service_factory,
)
from kodit.infrastructure.enricher.enricher_factory import (
    enricher_domain_service_factory,
)

# InMemoryGitTagRepository removed - now handled by InMemoryGitRepoRepository
from kodit.infrastructure.indexing.fusion_service import ReciprocalRankFusionService
from kodit.infrastructure.physical_architecture.formatters.narrative_formatter import (
    NarrativeFormatter,
)
from kodit.infrastructure.slicing.slicer import Slicer
from kodit.infrastructure.sqlalchemy.embedding_repository import (
    SqlAlchemyEmbeddingRepository,
    create_embedding_repository,
)
from kodit.infrastructure.sqlalchemy.enrichment_association_repository import (
    create_enrichment_association_repository,
)
from kodit.infrastructure.sqlalchemy.enrichment_v2_repository import (
    create_enrichment_v2_repository,
)
from kodit.infrastructure.sqlalchemy.git_branch_repository import (
    create_git_branch_repository,
)
from kodit.infrastructure.sqlalchemy.git_commit_repository import (
    create_git_commit_repository,
)
from kodit.infrastructure.sqlalchemy.git_file_repository import (
    create_git_file_repository,
)
from kodit.infrastructure.sqlalchemy.git_repository import create_git_repo_repository
from kodit.infrastructure.sqlalchemy.git_tag_repository import (
    create_git_tag_repository,
)
from kodit.infrastructure.sqlalchemy.task_status_repository import (
    create_task_status_repository,
)
from kodit.infrastructure.sqlalchemy.unit_of_work import SqlAlchemyUnitOfWork


class ServerFactory:
    """Factory for creating server application services."""

    def __init__(
        self,
        app_context: AppContext,
        session_factory: Callable[[], AsyncSession],
    ) -> None:
        """Initialize the ServerFactory."""
        self.app_context = app_context
        self.session_factory = session_factory
        self._repo_repository: GitRepoRepository | None = None
        self._snippet_v2_repository: SnippetRepositoryV2 | None = None
        self._git_adapter: GitAdapter | None = None
        self._scanner: GitRepositoryScanner | None = None
        self._cloner: RepositoryCloner | None = None
        self._commit_indexing_application_service: (
            CommitIndexingApplicationService | None
        ) = None
        self._enricher_service: Enricher | None = None
        self._task_status_repository: TaskStatusRepository | None = None
        self._operation: ProgressTracker | None = None
        self._queue_service: QueueService | None = None
        self._slicer: Slicer | None = None
        self._bm25_service: BM25DomainService | None = None
        self._bm25_repository: BM25Repository | None = None
        self._code_search_service: EmbeddingDomainService | None = None
        self._text_search_service: EmbeddingDomainService | None = None
        self._sync_scheduler_service: SyncSchedulerService | None = None
        self._embedding_repository: SqlAlchemyEmbeddingRepository | None = None
        self._fusion_service: FusionService | None = None
        self._code_search_application_service: CodeSearchApplicationService | None = (
            None
        )
        self._git_commit_repository: GitCommitRepository | None = None
        self._git_file_repository: GitFileRepository | None = None
        self._git_branch_repository: GitBranchRepository | None = None
        self._git_tag_repository: GitTagRepository | None = None
        self._architecture_service: PhysicalArchitectureService | None = None
        self._enrichment_v2_repository: EnrichmentV2Repository | None = None
        self._enrichment_association_repository: (
            EnrichmentAssociationRepository | None
        ) = None
        self._architecture_formatter: PhysicalArchitectureFormatter | None = None
        self._cookbook_context_service: CookbookContextService | None = None
        self._trackable_resolution_service: TrackableResolutionService | None = None
        self._enrichment_query_service: EnrichmentQueryService | None = None
        self._repository_query_service: RepositoryQueryService | None = None
        self._handler_registry: TaskHandlerRegistry | None = None
        self._repository_sync_service: RepositorySyncService | None = None

    def architecture_formatter(self) -> PhysicalArchitectureFormatter:
        """Create a PhysicalArchitectureFormatter instance."""
        if not self._architecture_formatter:
            self._architecture_formatter = NarrativeFormatter()
        return self._architecture_formatter

    def architecture_service(self) -> PhysicalArchitectureService:
        """Create a PhysicalArchitectureService instance."""
        if not self._architecture_service:
            self._architecture_service = PhysicalArchitectureService(
                formatter=self.architecture_formatter()
            )
        return self._architecture_service

    def cookbook_context_service(self) -> CookbookContextService:
        """Create a CookbookContextService instance."""
        if not self._cookbook_context_service:
            self._cookbook_context_service = CookbookContextService()
        return self._cookbook_context_service

    def enrichment_v2_repository(self) -> EnrichmentV2Repository:
        """Create a EnrichmentV2Repository instance."""
        if not self._enrichment_v2_repository:
            self._enrichment_v2_repository = create_enrichment_v2_repository(
                session_factory=self.session_factory
            )
        return self._enrichment_v2_repository

    def enrichment_association_repository(
        self,
    ) -> EnrichmentAssociationRepository:
        """Create a EnrichmentAssociationRepository instance."""
        if not self._enrichment_association_repository:
            self._enrichment_association_repository = (
                create_enrichment_association_repository(
                    session_factory=self.session_factory
                )
            )
        return self._enrichment_association_repository

    def queue_service(self) -> QueueService:
        """Create a QueueService instance."""
        if not self._queue_service:
            self._queue_service = QueueService(session_factory=self.session_factory)
        return self._queue_service

    def task_status_repository(self) -> TaskStatusRepository:
        """Create a TaskStatusRepository instance."""
        if not self._task_status_repository:
            self._task_status_repository = create_task_status_repository(
                session_factory=self.session_factory
            )
        return self._task_status_repository

    def operation(self) -> ProgressTracker:
        """Create a ProgressTracker instance."""
        if not self._operation:
            self._operation = create_server_operation(
                task_status_repository=self.task_status_repository()
            )
        return self._operation

    def slicer(self) -> Slicer:
        """Create a Slicer instance."""
        if not self._slicer:
            self._slicer = Slicer()
        return self._slicer

    def bm25_repository(self) -> BM25Repository:
        """Create a BM25Repository instance."""
        if not self._bm25_repository:
            if self.app_context.default_search.provider == "vectorchord":
                self._bm25_repository = VectorChordBM25Repository(
                    session_factory=self.session_factory
                )
            else:
                self._bm25_repository = LocalBM25Repository(
                    data_dir=self.app_context.get_data_dir()
                )
        return self._bm25_repository

    def bm25_service(self) -> BM25DomainService:
        """Create a BM25DomainService instance."""
        if not self._bm25_service:
            self._bm25_service = BM25DomainService(repository=self.bm25_repository())
        return self._bm25_service

    def code_search_service(self) -> EmbeddingDomainService:
        """Create a EmbeddingDomainService instance."""
        if not self._code_search_service:
            self._code_search_service = embedding_domain_service_factory(
                "code", self.app_context, self.session_factory
            )
        return self._code_search_service

    def text_search_service(self) -> EmbeddingDomainService:
        """Create a EmbeddingDomainService instance."""
        if not self._text_search_service:
            self._text_search_service = embedding_domain_service_factory(
                "text", self.app_context, self.session_factory
            )
        return self._text_search_service

    def repository_sync_service(self) -> RepositorySyncService:
        """Create a RepositorySyncService instance."""
        if not self._repository_sync_service:
            self._repository_sync_service = RepositorySyncService(
                scanner=self.scanner(),
                git_commit_repository=self.git_commit_repository(),
                git_branch_repository=self.git_branch_repository(),
                git_tag_repository=self.git_tag_repository(),
            )
        return self._repository_sync_service

    def handler_registry(self) -> TaskHandlerRegistry:
        """Create and populate a TaskHandlerRegistry instance."""
        if not self._handler_registry:
            from kodit.domain.value_objects import TaskOperation

            registry = TaskHandlerRegistry()

            # Register commit operation handlers
            registry.register(
                TaskOperation.SCAN_COMMIT,
                ScanCommitHandler(
                    repo_repository=self.repo_repository(),
                    git_commit_repository=self.git_commit_repository(),
                    git_file_repository=self.git_file_repository(),
                    scanner=self.scanner(),
                    operation=self.operation(),
                ),
            )
            registry.register(
                TaskOperation.EXTRACT_SNIPPETS_FOR_COMMIT,
                ExtractSnippetsHandler(
                    repo_repository=self.repo_repository(),
                    git_commit_repository=self.git_commit_repository(),
                    scanner=self.scanner(),
                    enrichment_v2_repository=self.enrichment_v2_repository(),
                    enrichment_association_repository=self.enrichment_association_repository(),
                    enrichment_query_service=self.enrichment_query_service(),
                    operation=self.operation(),
                ),
            )
            registry.register(
                TaskOperation.EXTRACT_EXAMPLES_FOR_COMMIT,
                ExtractExamplesHandler(
                    repo_repository=self.repo_repository(),
                    git_commit_repository=self.git_commit_repository(),
                    scanner=self.scanner(),
                    enrichment_v2_repository=self.enrichment_v2_repository(),
                    enrichment_association_repository=self.enrichment_association_repository(),
                    enrichment_query_service=self.enrichment_query_service(),
                    operation=self.operation(),
                ),
            )
            registry.register(
                TaskOperation.CREATE_BM25_INDEX_FOR_COMMIT,
                CreateBM25IndexHandler(
                    bm25_service=self.bm25_service(),
                    enrichment_query_service=self.enrichment_query_service(),
                    operation=self.operation(),
                ),
            )
            registry.register(
                TaskOperation.CREATE_CODE_EMBEDDINGS_FOR_COMMIT,
                CreateCodeEmbeddingsHandler(
                    code_search_service=self.code_search_service(),
                    embedding_repository=self.embedding_repository(),
                    enrichment_query_service=self.enrichment_query_service(),
                    operation=self.operation(),
                ),
            )
            registry.register(
                TaskOperation.CREATE_EXAMPLE_CODE_EMBEDDINGS_FOR_COMMIT,
                CreateExampleCodeEmbeddingsHandler(
                    code_search_service=self.code_search_service(),
                    embedding_repository=self.embedding_repository(),
                    enrichment_query_service=self.enrichment_query_service(),
                    operation=self.operation(),
                ),
            )
            registry.register(
                TaskOperation.CREATE_SUMMARY_ENRICHMENT_FOR_COMMIT,
                CreateSummaryEnrichmentHandler(
                    enricher_service=self.enricher(),
                    enrichment_v2_repository=self.enrichment_v2_repository(),
                    enrichment_association_repository=self.enrichment_association_repository(),
                    enrichment_query_service=self.enrichment_query_service(),
                    operation=self.operation(),
                ),
            )
            registry.register(
                TaskOperation.CREATE_EXAMPLE_SUMMARY_FOR_COMMIT,
                CreateExampleSummaryHandler(
                    enricher_service=self.enricher(),
                    enrichment_v2_repository=self.enrichment_v2_repository(),
                    enrichment_association_repository=self.enrichment_association_repository(),
                    enrichment_query_service=self.enrichment_query_service(),
                    operation=self.operation(),
                ),
            )
            registry.register(
                TaskOperation.CREATE_SUMMARY_EMBEDDINGS_FOR_COMMIT,
                CreateSummaryEmbeddingsHandler(
                    text_search_service=self.text_search_service(),
                    embedding_repository=self.embedding_repository(),
                    enrichment_v2_repository=self.enrichment_v2_repository(),
                    enrichment_association_repository=self.enrichment_association_repository(),
                    enrichment_query_service=self.enrichment_query_service(),
                    operation=self.operation(),
                ),
            )
            registry.register(
                TaskOperation.CREATE_EXAMPLE_SUMMARY_EMBEDDINGS_FOR_COMMIT,
                CreateExampleSummaryEmbeddingsHandler(
                    text_search_service=self.text_search_service(),
                    embedding_repository=self.embedding_repository(),
                    enrichment_v2_repository=self.enrichment_v2_repository(),
                    enrichment_association_repository=self.enrichment_association_repository(),
                    enrichment_query_service=self.enrichment_query_service(),
                    operation=self.operation(),
                ),
            )
            registry.register(
                TaskOperation.CREATE_ARCHITECTURE_ENRICHMENT_FOR_COMMIT,
                ArchitectureDiscoveryHandler(
                    repo_repository=self.repo_repository(),
                    architecture_service=self.architecture_service(),
                    enricher_service=self.enricher(),
                    enrichment_v2_repository=self.enrichment_v2_repository(),
                    enrichment_association_repository=self.enrichment_association_repository(),
                    enrichment_query_service=self.enrichment_query_service(),
                    operation=self.operation(),
                ),
            )
            registry.register(
                TaskOperation.CREATE_PUBLIC_API_DOCS_FOR_COMMIT,
                ApiDocsHandler(
                    repo_repository=self.repo_repository(),
                    git_file_repository=self.git_file_repository(),
                    enrichment_v2_repository=self.enrichment_v2_repository(),
                    enrichment_association_repository=self.enrichment_association_repository(),
                    enrichment_query_service=self.enrichment_query_service(),
                    operation=self.operation(),
                ),
            )
            registry.register(
                TaskOperation.CREATE_COMMIT_DESCRIPTION_FOR_COMMIT,
                CommitDescriptionHandler(
                    repo_repository=self.repo_repository(),
                    scanner=self.scanner(),
                    enricher_service=self.enricher(),
                    enrichment_v2_repository=self.enrichment_v2_repository(),
                    enrichment_association_repository=self.enrichment_association_repository(),
                    enrichment_query_service=self.enrichment_query_service(),
                    operation=self.operation(),
                ),
            )
            registry.register(
                TaskOperation.CREATE_DATABASE_SCHEMA_FOR_COMMIT,
                DatabaseSchemaHandler(
                    repo_repository=self.repo_repository(),
                    database_schema_detector=DatabaseSchemaDetector(),
                    enricher_service=self.enricher(),
                    enrichment_v2_repository=self.enrichment_v2_repository(),
                    enrichment_association_repository=self.enrichment_association_repository(),
                    enrichment_query_service=self.enrichment_query_service(),
                    operation=self.operation(),
                ),
            )
            registry.register(
                TaskOperation.CREATE_COOKBOOK_FOR_COMMIT,
                CookbookHandler(
                    repo_repository=self.repo_repository(),
                    git_file_repository=self.git_file_repository(),
                    cookbook_context_service=self.cookbook_context_service(),
                    enricher_service=self.enricher(),
                    enrichment_v2_repository=self.enrichment_v2_repository(),
                    enrichment_association_repository=self.enrichment_association_repository(),
                    enrichment_query_service=self.enrichment_query_service(),
                    operation=self.operation(),
                ),
            )

            # Register repository operation handlers
            registry.register(
                TaskOperation.CLONE_REPOSITORY,
                CloneRepositoryHandler(
                    repo_repository=self.repo_repository(),
                    cloner=self.cloner(),
                    repository_sync_service=self.repository_sync_service(),
                    repository_query_service=self.repository_query_service(),
                    queue=self.queue_service(),
                    operation=self.operation(),
                ),
            )
            registry.register(
                TaskOperation.SYNC_REPOSITORY,
                SyncRepositoryHandler(
                    repo_repository=self.repo_repository(),
                    git_commit_repository=self.git_commit_repository(),
                    cloner=self.cloner(),
                    repository_sync_service=self.repository_sync_service(),
                    repository_query_service=self.repository_query_service(),
                    queue=self.queue_service(),
                    operation=self.operation(),
                ),
            )
            registry.register(
                TaskOperation.DELETE_REPOSITORY,
                DeleteRepositoryHandler(
                    repo_repository=self.repo_repository(),
                    git_commit_repository=self.git_commit_repository(),
                    git_file_repository=self.git_file_repository(),
                    git_branch_repository=self.git_branch_repository(),
                    git_tag_repository=self.git_tag_repository(),
                    bm25_service=self.bm25_service(),
                    embedding_repository=self.embedding_repository(),
                    enrichment_v2_repository=self.enrichment_v2_repository(),
                    enrichment_association_repository=self.enrichment_association_repository(),
                    enrichment_query_service=self.enrichment_query_service(),
                    operation=self.operation(),
                ),
            )
            registry.register(
                TaskOperation.RESCAN_COMMIT,
                RescanCommitHandler(
                    git_file_repository=self.git_file_repository(),
                    bm25_service=self.bm25_service(),
                    embedding_repository=self.embedding_repository(),
                    enrichment_v2_repository=self.enrichment_v2_repository(),
                    enrichment_association_repository=self.enrichment_association_repository(),
                    enrichment_query_service=self.enrichment_query_service(),
                    queue=self.queue_service(),
                    operation=self.operation(),
                ),
            )

            self._handler_registry = registry
        return self._handler_registry

    def commit_indexing_application_service(self) -> CommitIndexingApplicationService:
        """Create a CommitIndexingApplicationService instance."""
        if not self._commit_indexing_application_service:
            self._commit_indexing_application_service = (
                CommitIndexingApplicationService(
                    repo_repository=self.repo_repository(),
                    task_status_repository=self.task_status_repository(),
                    operation=self.operation(),
                    queue=self.queue_service(),
                    handler_registry=self.handler_registry(),
                )
            )

        return self._commit_indexing_application_service

    def unit_of_work(self) -> SqlAlchemyUnitOfWork:
        """Create a SqlAlchemyUnitOfWork instance."""
        return SqlAlchemyUnitOfWork(session_factory=self.session_factory)

    def repo_repository(self) -> GitRepoRepository:
        """Create a GitRepoRepository instance."""
        if not self._repo_repository:
            self._repo_repository = create_git_repo_repository(
                session_factory=self.session_factory
            )
        return self._repo_repository

    # branch_repository and commit_repository removed - now handled by repo_repository
    # as GitRepo is the aggregate root

    def git_adapter(self) -> GitAdapter:
        """Create a GitAdapter instance."""
        if not self._git_adapter:
            self._git_adapter = create_git_adapter(self.app_context.git.provider)
        return self._git_adapter

    # tag_repository removed - now handled by repo_repository

    def scanner(self) -> GitRepositoryScanner:
        """Create a GitRepositoryScanner instance."""
        if not self._scanner:
            self._scanner = GitRepositoryScanner(self.git_adapter())
        return self._scanner

    def cloner(self) -> RepositoryCloner:
        """Create a RepositoryCloner instance."""
        if not self._cloner:
            self._cloner = RepositoryCloner(
                self.git_adapter(), self.app_context.get_clone_dir()
            )
        return self._cloner

    def git_file_repository(self) -> GitFileRepository:
        """Create a GitFileRepository instance."""
        if not self._git_file_repository:
            self._git_file_repository = create_git_file_repository(
                session_factory=self.session_factory
            )
        return self._git_file_repository

    def enricher(self) -> Enricher:
        """Create a EnricherDomainService instance."""
        if not self._enricher_service:
            self._enricher_service = enricher_domain_service_factory(self.app_context)
        return self._enricher_service

    def sync_scheduler_service(self) -> SyncSchedulerService:
        """Create a SyncSchedulerService instance."""
        if not self._sync_scheduler_service:
            self._sync_scheduler_service = SyncSchedulerService(
                queue_service=self.queue_service(),
                repo_repository=self.repo_repository(),
                task_status_repository=self.task_status_repository(),
            )
        return self._sync_scheduler_service

    def embedding_repository(self) -> SqlAlchemyEmbeddingRepository:
        """Create a SqlAlchemyEmbeddingRepository instance."""
        if not self._embedding_repository:
            self._embedding_repository = create_embedding_repository(
                session_factory=self.session_factory
            )
        return self._embedding_repository

    def fusion_service(self) -> FusionService:
        """Create a FusionService instance."""
        if not self._fusion_service:
            self._fusion_service = ReciprocalRankFusionService()
        return self._fusion_service

    def code_search_application_service(self) -> CodeSearchApplicationService:
        """Create a CodeSearchApplicationService instance."""
        if not self._code_search_application_service:
            self._code_search_application_service = CodeSearchApplicationService(
                bm25_service=self.bm25_service(),
                code_search_service=self.code_search_service(),
                text_search_service=self.text_search_service(),
                progress_tracker=self.operation(),
                fusion_service=self.fusion_service(),
                enrichment_query_service=self.enrichment_query_service(),
            )
        return self._code_search_application_service

    def git_commit_repository(self) -> GitCommitRepository:
        """Create a GitCommitRepository instance."""
        if not self._git_commit_repository:
            self._git_commit_repository = create_git_commit_repository(
                session_factory=self.session_factory
            )
        return self._git_commit_repository

    def git_branch_repository(self) -> GitBranchRepository:
        """Create a GitBranchRepository instance."""
        if not self._git_branch_repository:
            self._git_branch_repository = create_git_branch_repository(
                session_factory=self.session_factory
            )
        return self._git_branch_repository

    def git_tag_repository(self) -> GitTagRepository:
        """Create a GitTagRepository instance."""
        if not self._git_tag_repository:
            self._git_tag_repository = create_git_tag_repository(
                session_factory=self.session_factory
            )
        return self._git_tag_repository

    def trackable_resolution_service(self) -> TrackableResolutionService:
        """Create a TrackableResolutionService instance."""
        if not self._trackable_resolution_service:
            self._trackable_resolution_service = TrackableResolutionService(
                commit_repo=self.git_commit_repository(),
                branch_repo=self.git_branch_repository(),
                tag_repo=self.git_tag_repository(),
            )
        return self._trackable_resolution_service

    def enrichment_query_service(self) -> EnrichmentQueryService:
        """Create a EnrichmentQueryService instance."""
        if not self._enrichment_query_service:
            self._enrichment_query_service = EnrichmentQueryService(
                trackable_resolution=self.trackable_resolution_service(),
                enrichment_repo=self.enrichment_v2_repository(),
                enrichment_association_repository=self.enrichment_association_repository(),
            )
        return self._enrichment_query_service

    def repository_query_service(self) -> RepositoryQueryService:
        """Create a RepositoryQueryService instance."""
        if not self._repository_query_service:
            self._repository_query_service = RepositoryQueryService(
                git_repo_repository=self.repo_repository(),
                trackable_resolution=self.trackable_resolution_service(),
                git_adapter=self.git_adapter(),
            )
        return self._repository_query_service



#!/usr/bin/env python3
"""Convert OpenAPI 3.1 spec to 3.0 for better tool compatibility."""

import json
import sys
from pathlib import Path


def convert_31_to_30(spec: dict) -> dict:
    """Convert OpenAPI 3.1 spec to 3.0.

    This handles the main compatibility issues:
    - Change version from 3.1.x to 3.0.3
    - Convert nullable types from type arrays to nullable property
    """
    spec = spec.copy()

    # Update version
    spec["openapi"] = "3.0.3"

    # Process components
    if "components" in spec and "schemas" in spec["components"]:
        spec["components"]["schemas"] = convert_schemas(spec["components"]["schemas"])

    # Process paths
    if "paths" in spec:
        spec["paths"] = convert_paths(spec["paths"])

    return spec


def convert_schemas(schemas: dict) -> dict:
    """Convert schema definitions."""
    result = {}
    for name, schema in schemas.items():
        result[name] = convert_schema(schema)
    return result


def _handle_type_array(schema: dict) -> dict:
    """Handle type arrays with null (OpenAPI 3.1 feature)."""
    if "type" not in schema or not isinstance(schema["type"], list):
        return schema

    types = schema["type"]
    if "null" not in types:
        return schema

    schema["nullable"] = True
    non_null_types = [t for t in types if t != "null"]
    if len(non_null_types) == 1:
        schema["type"] = non_null_types[0]
    elif non_null_types:
        schema.pop("type")
        schema["anyOf"] = [{"type": t} for t in non_null_types]

    return schema


def _handle_any_of(schema: dict) -> dict:
    """Handle anyOf with null."""
    if "anyOf" not in schema:
        return schema

    any_of = schema["anyOf"]
    null_schemas = [s for s in any_of if s.get("type") == "null"]
    non_null_schemas = [s for s in any_of if s.get("type") != "null"]

    if not null_schemas:
        return schema

    schema["nullable"] = True
    if non_null_schemas:
        schema["anyOf"] = [convert_schema(s) for s in non_null_schemas]
    else:
        schema.pop("anyOf", None)

    return schema


def convert_schema(schema: dict | list) -> dict | list:
    """Convert a single schema, handling nullable types."""
    if isinstance(schema, list):
        return [convert_schema(s) for s in schema]

    if not isinstance(schema, dict):
        return schema

    schema = schema.copy()

    schema = _handle_type_array(schema)
    schema = _handle_any_of(schema)

    # Recursively process nested schemas
    for key in ["properties", "items", "additionalProperties"]:
        if key in schema:
            if key == "properties":
                schema[key] = {k: convert_schema(v) for k, v in schema[key].items()}
            else:
                schema[key] = convert_schema(schema[key])

    if "allOf" in schema:
        schema["allOf"] = [convert_schema(s) for s in schema["allOf"]]

    return schema


def convert_paths(paths: dict) -> dict:
    """Convert path definitions."""
    result = {}
    for path, path_item in paths.items():
        result[path] = convert_path_item(path_item)
    return result


def convert_path_item(path_item: dict) -> dict:
    """Convert a path item."""
    path_item = path_item.copy()

    for method in ["get", "post", "put", "delete", "patch", "options", "head", "trace"]:
        if method in path_item:
            path_item[method] = convert_operation(path_item[method])

    return path_item


def convert_operation(operation: dict) -> dict:
    """Convert an operation."""
    operation = operation.copy()

    # Convert parameters
    if "parameters" in operation:
        operation["parameters"] = [
            convert_parameter(p) for p in operation["parameters"]
        ]

    # Convert request body
    if "requestBody" in operation:
        operation["requestBody"] = convert_request_body(operation["requestBody"])

    # Convert responses
    if "responses" in operation:
        operation["responses"] = {
            status: convert_response(resp)
            for status, resp in operation["responses"].items()
        }

    return operation


def convert_parameter(param: dict) -> dict:
    """Convert a parameter."""
    param = param.copy()
    if "schema" in param:
        param["schema"] = convert_schema(param["schema"])
    return param


def convert_request_body(request_body: dict) -> dict:
    """Convert a request body."""
    request_body = request_body.copy()
    if "content" in request_body:
        request_body["content"] = convert_content(request_body["content"])
    return request_body


def convert_response(response: dict) -> dict:
    """Convert a response."""
    response = response.copy()
    if "content" in response:
        response["content"] = convert_content(response["content"])
    return response


def convert_content(content: dict) -> dict:
    """Convert content definitions."""
    result = {}
    for media_type, media_type_obj in content.items():
        converted_obj = media_type_obj.copy()
        if "schema" in converted_obj:
            converted_obj["schema"] = convert_schema(converted_obj["schema"])
        result[media_type] = converted_obj
    return result


def main() -> None:
    """Convert OpenAPI file from 3.1 to 3.0 format."""
    if len(sys.argv) != 3:
        sys.exit(1)

    input_file = Path(sys.argv[1])
    output_file = Path(sys.argv[2])

    with input_file.open() as f:
        spec = json.load(f)

    converted = convert_31_to_30(spec)

    with output_file.open("w") as f:
        json.dump(converted, f, indent=2)



if __name__ == "__main__":
    main()



"""Commit JSON-API schemas."""

from datetime import datetime

from pydantic import BaseModel


class GitFileData(BaseModel):
    """Git file data."""

    blob_sha: str
    path: str
    mime_type: str
    size: int


class CommitAttributes(BaseModel):
    """Commit attributes following JSON-API spec."""

    commit_sha: str
    date: datetime
    message: str
    parent_commit_sha: str
    author: str


class CommitData(BaseModel):
    """Commit data following JSON-API spec."""

    type: str = "commit"
    id: str
    attributes: CommitAttributes


class CommitResponse(BaseModel):
    """Single commit response following JSON-API spec."""

    data: CommitData


class CommitListResponse(BaseModel):
    """Commit list response following JSON-API spec."""

    data: list[CommitData]


class FileAttributes(BaseModel):
    """File attributes following JSON-API spec."""

    blob_sha: str
    path: str
    mime_type: str
    size: int
    extension: str


class FileData(BaseModel):
    """File data following JSON-API spec."""

    type: str = "file"
    id: str
    attributes: FileAttributes


class FileResponse(BaseModel):
    """Single file response following JSON-API spec."""

    data: FileData


class FileListResponse(BaseModel):
    """File list response following JSON-API spec."""

    data: list[FileData]


class EmbeddingAttributes(BaseModel):
    """Embedding attributes following JSON-API spec."""

    snippet_sha: str
    embedding_type: str
    embedding: list[float]


class EmbeddingData(BaseModel):
    """Embedding data following JSON-API spec."""

    type: str = "embedding"
    id: str
    attributes: EmbeddingAttributes


class EmbeddingListResponse(BaseModel):
    """Embedding list response following JSON-API spec."""

    data: list[EmbeddingData]



"""Commit description enrichment domain entity."""

from dataclasses import dataclass

from kodit.domain.enrichments.history.history import HistoryEnrichment

ENRICHMENT_SUBTYPE_COMMIT_DESCRIPTION = "commit_description"


@dataclass(frozen=True)
class CommitDescriptionEnrichment(HistoryEnrichment):
    """Enrichment containing a description of what a commit did."""

    @property
    def subtype(self) -> str | None:
        """Return the enrichment subtype."""
        return ENRICHMENT_SUBTYPE_COMMIT_DESCRIPTION



#!/usr/bin/env python3
# /// script
# dependencies = [
#   "httpx",
#   "structlog",
#   "aiosqlite",
# ]
# ///
"""Smoke test for repository deletion functionality.

This test verifies that when a repository is deleted via the API,
all associated data is properly removed from the database.
"""

import os
import socket
import subprocess
import sys
import tempfile
import time
from pathlib import Path

import httpx
import structlog

BASE_HOST = "127.0.0.1"
BASE_PORT = 8081
BASE_URL = f"http://{BASE_HOST}:{BASE_PORT}"
TARGET_URI = "https://gist.github.com/7aa38185e20433c04c533f2b28f4e217.git"

log = structlog.get_logger(__name__)


def is_port_available(host: str, port: int) -> bool:
    """Check if a port is available by trying to bind to it."""
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        try:
            s.bind((host, port))
        except OSError:
            return False  # Port is in use
        else:
            return True


def retry_with_timeout(
    func,  # noqa: ANN001
    timeout: float = 600,
    retry_delay: float = 1,
) -> None:
    """Keep trying a function until it succeeds or timeout is reached."""
    start_time = time.time()

    while time.time() - start_time < timeout:
        try:
            if func():
                return
        except Exception as e:
            if time.time() - start_time >= timeout:
                raise TimeoutError(f"Failed after {timeout} seconds: {e}") from e

        time.sleep(retry_delay)

    raise TimeoutError(f"Timed out after {timeout} seconds")


def verify_database_empty(db_path: str) -> None:
    """Verify that all repository-related tables are empty using sqlite3."""
    import sqlite3

    tables_to_check = [
        "git_repos",
        "git_commits",
        "git_branches",
        "git_tags",
        "git_commit_files",
        "embeddings",
        "enrichments_v2",
        "enrichment_associations",
    ]

    conn = sqlite3.connect(db_path)
    cursor = conn.cursor()

    try:
        for table in tables_to_check:
            cursor.execute(f"SELECT COUNT(*) FROM {table}")  # noqa: S608
            count = cursor.fetchone()[0]
            log.info("Checked table after deletion", table=table, count=count)
            if count > 0:
                log.error("Table not empty after deletion", table=table, count=count)
                msg = f"Table {table} still has {count} rows after deletion"
                raise AssertionError(msg)

        log.info("All tables are empty after deletion")
    finally:
        conn.close()


def main() -> None:  # noqa: PLR0915
    """Run smoke test for repository deletion."""
    # Check there is nothing running on the port already
    if not is_port_available(BASE_HOST, BASE_PORT):
        log.error("Port is already in use", host=BASE_HOST, port=BASE_PORT)
        sys.exit(1)

    # Create a temporary database file
    db_fd, db_path = tempfile.mkstemp(suffix=".db")
    os.close(db_fd)
    db_path_obj = Path(db_path)

    # Create a temporary file for the environment file
    with tempfile.NamedTemporaryFile(delete=False) as f:
        tmpfile = Path(f.name)

    env = os.environ.copy()
    env.update({
        "DISABLE_TELEMETRY": "true",
        "DB_URL": f"sqlite+aiosqlite:///{db_path}",
    })

    prefix = [] if os.getenv("CI") else ["uv", "run"]
    cmd = [
        *prefix,
        "kodit",
        "--env-file",
        str(tmpfile),
        "serve",
        "--host",
        BASE_HOST,
        "--port",
        str(BASE_PORT),
    ]
    process = subprocess.Popen(cmd, env=env)  # noqa: S603

    try:
        with httpx.Client(timeout=30.0) as client:
            log.info("Waiting for server to start listening", port=BASE_PORT)
            while is_port_available(BASE_HOST, BASE_PORT):
                time.sleep(1)
            client.get(f"{BASE_URL}/healthz").raise_for_status()  # cspell: disable-line
            log.info("Server health check passed")

            # Create a repository
            log.info("Creating repository", uri=TARGET_URI)
            payload = {
                "data": {"type": "repository", "attributes": {"remote_uri": TARGET_URI}}
            }
            client.post(
                f"{BASE_URL}/api/v1/repositories", json=payload
            ).raise_for_status()
            log.info("Repository created successfully")

            # Get the repository ID  # cspell: disable-line
            response = client.get(f"{BASE_URL}/api/v1/repositories")
            repos = response.json()
            repo_id = repos["data"][0]["id"]
            log.info("Retrieved repository ID", repo_id=repo_id)

            # Wait for indexing to complete
            log.info("Waiting for indexing to complete", repo_id=repo_id)

            def indexing_finished() -> bool:
                response = client.get(
                    f"{BASE_URL}/api/v1/repositories/{repo_id}/status"
                )
                status = response.json()
                return (
                    all(
                        task["attributes"]["state"] == "completed"
                        or task["attributes"]["state"] == "skipped"
                        for task in status["data"]
                    )
                    and len(status["data"]) > 5
                )

            retry_with_timeout(indexing_finished)
            log.info("Indexing completed", repo_id=repo_id)

            # Verify data exists before deletion
            log.info("Verifying data exists before deletion")
            import sqlite3
            conn = sqlite3.connect(db_path)
            cursor = conn.cursor()
            cursor.execute("SELECT COUNT(*) FROM git_repos")
            repo_count = cursor.fetchone()[0]
            cursor.execute("SELECT COUNT(*) FROM git_commits")
            commit_count = cursor.fetchone()[0]
            cursor.execute("SELECT COUNT(*) FROM git_branches")
            branch_count = cursor.fetchone()[0]
            conn.close()

            log.info(
                "Database state before deletion",
                repos=repo_count,
                commits=commit_count,
                branches=branch_count,
            )

            if repo_count == 0:
                log.error("No repository found before deletion")
                raise AssertionError("No repository found before deletion")

            # Delete the repository
            log.info("Deleting repository", repo_id=repo_id)
            client.delete(
                f"{BASE_URL}/api/v1/repositories/{repo_id}"
            ).raise_for_status()
            log.info("Repository deleted successfully", repo_id=repo_id)

            # Verify all data is deleted
            log.info("Verifying all data is deleted from database")
            verify_database_empty(db_path)

            log.info("Deletion smoke test passed successfully")

    finally:
        process.terminate()
        process.wait(timeout=10)
        tmpfile.unlink(missing_ok=True)
        db_path_obj.unlink(missing_ok=True)


if __name__ == "__main__":
    main()



"""Enricher factory for creating generic enricher domain services."""

from kodit.config import AppContext, Endpoint
from kodit.domain.enrichments.enricher import Enricher
from kodit.infrastructure.enricher.litellm_enricher import LiteLLMEnricher
from kodit.infrastructure.enricher.local_enricher import LocalEnricher
from kodit.infrastructure.enricher.null_enricher import NullEnricher
from kodit.log import log_event


def _get_endpoint_configuration(app_context: AppContext) -> Endpoint | None:
    """Get the endpoint configuration for the enricher service.

    Args:
        app_context: The application context.

    Returns:
        The endpoint configuration or None.

    """
    return app_context.enrichment_endpoint or None


def enricher_domain_service_factory(
    app_context: AppContext,
    *,
    use_null_enricher: bool = False,
) -> Enricher:
    """Create an enricher domain service.

    Args:
        app_context: The application context.
        use_null_enricher: Whether to use the null enricher instead.

    Returns:
        An enricher domain service instance.

    """
    enricher: Enricher

    if use_null_enricher:
        log_event("kodit.enricher", {"provider": "null"})
        enricher = NullEnricher()
    else:
        endpoint = _get_endpoint_configuration(app_context)
        if endpoint:
            log_event("kodit.enricher", {"provider": "litellm"})
            enricher = LiteLLMEnricher(endpoint=endpoint)
        else:
            log_event("kodit.enricher", {"provider": "local"})
            enricher = LocalEnricher()

    return enricher



"""Architecture enrichment domain entity."""

from abc import ABC
from dataclasses import dataclass

from kodit.domain.enrichments.enrichment import (
    CommitEnrichment,
)

ENRICHMENT_TYPE_ARCHITECTURE = "architecture"


@dataclass(frozen=True)
class ArchitectureEnrichment(CommitEnrichment, ABC):
    """Enrichment containing physical architecture discovery for a commit."""

    @property
    def type(self) -> str:
        """Return the enrichment type."""
        return ENRICHMENT_TYPE_ARCHITECTURE



"""Narrative formatter for converting observations to LLM-optimized text."""

import re

from kodit.domain.enrichments.architecture.physical.discovery_notes import (
    ArchitectureDiscoveryNotes,
)


class NarrativeFormatter:
    """Formats architecture observations into narrative text optimized for LLM consumption."""  # noqa: E501

    def format_for_llm(self, notes: ArchitectureDiscoveryNotes) -> str:
        """Convert discovery notes into a comprehensive narrative format."""
        sections = []

        # Title and overview
        sections.append("# Physical Architecture Discovery Report")
        sections.append("")
        sections.append(notes.repository_context)
        sections.append("")

        # Component Analysis
        self._add_component_section(sections, notes.component_observations)

        # Connection Analysis
        self._add_connection_section(sections, notes.connection_observations)

        # Infrastructure Analysis
        self._add_infrastructure_section(sections, notes.infrastructure_observations)

        # Methodology
        sections.append("## Discovery Methodology")
        sections.append(notes.discovery_metadata)
        sections.append("")

        # Conclusion
        self._add_conclusion_section(sections, notes)

        return "\n".join(sections)

    def _add_component_section(
        self, sections: list[str], component_observations: list[str]
    ) -> None:
        """Add component observations section with proper formatting."""
        sections.append("## Components")
        sections.append("")

        if component_observations:
            for i, observation in enumerate(component_observations, 1):
                sections.append(f"**{i}.** {observation}")
            sections.append("")

            # Extract and highlight port information
            port_info = self._extract_port_information(component_observations)
            if port_info:
                sections.append("### Port Mappings")
                sections.append("")
                for component, ports_desc in port_info.items():
                    sections.append(f"- **{component}**: {ports_desc}")
                sections.append("")
        else:
            sections.append("None. Likely monolithic or library architecture.")
            sections.append("")

    def _add_connection_section(
        self, sections: list[str], connection_observations: list[str]
    ) -> None:
        """Add connection observations section with proper formatting."""
        sections.append("## Connections")
        sections.append("")

        if connection_observations:
            for i, observation in enumerate(connection_observations, 1):
                sections.append(f"**{i}.** {observation}")
            sections.append("")
        else:
            sections.append("None. Possible monolithic or independent services.")
            sections.append("")

    def _add_infrastructure_section(
        self, sections: list[str], infrastructure_observations: list[str]
    ) -> None:
        """Add infrastructure observations section with proper formatting."""
        sections.append("## Infrastructure")
        sections.append("")

        if infrastructure_observations:
            for i, observation in enumerate(infrastructure_observations, 1):
                sections.append(f"**{i}.** {observation}")
            sections.append("")
        else:
            sections.append("None. May use external or cloud-native deployment.")
            sections.append("")

    def _add_conclusion_section(
        self, sections: list[str], notes: ArchitectureDiscoveryNotes
    ) -> None:
        """Add a conclusion section summarizing the findings."""
        sections.append("## Summary")
        sections.append("")

        # Determine architecture characteristics
        has_components = bool(notes.component_observations)
        has_connections = bool(notes.connection_observations)
        has_infrastructure = bool(notes.infrastructure_observations)

        if has_components and has_connections and has_infrastructure:
            arch_type = "distributed microservices"
            complexity = "high"
        elif has_components and (has_connections or has_infrastructure):
            arch_type = "multi-component"
            complexity = "medium"
        elif has_components or has_infrastructure:
            arch_type = "structured application"
            complexity = "medium"
        else:
            arch_type = "monolithic"
            complexity = "low"

        sections.append(f"**Architecture:** {arch_type} | **Complexity:** {complexity}")
        sections.append("")
        sections.append("**Note:** Static analysis only. Runtime behavior may differ.")

    def _extract_port_information(
        self, component_observations: list[str]
    ) -> dict[str, str]:
        """Extract port information from component observations."""
        port_info = {}

        # Pattern to extract service name and port information
        service_pattern = r"Found '([^']+)' service"
        port_pattern = r"Exposes ports ([\d, ]+)(?: suggesting ([^.]+))?"

        for observation in component_observations:
            service_match = re.search(service_pattern, observation)
            port_match = re.search(port_pattern, observation)

            if service_match and port_match:
                service_name = service_match.group(1)
                ports = port_match.group(1).strip()
                protocol = port_match.group(2)

                if protocol:
                    port_info[service_name] = f"{ports} ({protocol})"
                else:
                    port_info[service_name] = ports

        return port_info



"""Domain service for resolving trackables to commits."""

import structlog

from kodit.domain.protocols import (
    GitBranchRepository,
    GitCommitRepository,
    GitTagRepository,
)
from kodit.domain.tracking.trackable import Trackable, TrackableReferenceType


class TrackableResolutionService:
    """Resolves trackables to ordered lists of commits.

    This is a domain service because it orchestrates multiple aggregates
    (branches, tags, commits) without belonging to any single entity.
    """

    def __init__(
        self,
        commit_repo: GitCommitRepository,
        branch_repo: GitBranchRepository,
        tag_repo: GitTagRepository,
    ) -> None:
        """Initialize the trackable resolution service."""
        self.commit_repo = commit_repo
        self.branch_repo = branch_repo
        self.tag_repo = tag_repo
        self.log = structlog.get_logger(__name__)

    async def resolve_to_commits(
        self, trackable: Trackable, limit: int = 100
    ) -> list[str]:
        """Resolve a trackable to an ordered list of commit SHAs.

        Returns commits from newest to oldest based on git history.
        """
        if trackable.type == TrackableReferenceType.BRANCH:
            return await self._resolve_branch(trackable, limit)
        if trackable.type == TrackableReferenceType.TAG:
            return await self._resolve_tag(trackable, limit)
        # COMMIT_SHA
        return [trackable.identifier]

    async def _resolve_branch(self, trackable: Trackable, limit: int) -> list[str]:
        """Get commits from branch HEAD backwards through history."""
        branch = await self.branch_repo.get_by_name(
            trackable.identifier, trackable.repo_id
        )
        # Walk commit history from head_commit backwards
        return await self._walk_commit_history(branch.head_commit_sha, limit)

    async def _resolve_tag(self, trackable: Trackable, limit: int) -> list[str]:
        """Get commits from tag target backwards through history."""
        tag = await self.tag_repo.get_by_name(trackable.identifier, trackable.repo_id)
        return await self._walk_commit_history(tag.target_commit_sha, limit)

    async def _walk_commit_history(self, start_sha: str, limit: int) -> list[str]:
        """Walk commit history backwards from start_sha."""
        result = []
        current_sha: str | None = start_sha

        for _ in range(limit):
            if not current_sha:
                break
            result.append(current_sha)
            if await self.commit_repo.exists(current_sha):
                commit = await self.commit_repo.get(current_sha)
                current_sha = commit.parent_commit_sha or None
            else:
                current_sha = None

        return result



"""Local BM25 repository implementation."""

from __future__ import annotations

import json
from pathlib import Path
from typing import TYPE_CHECKING

import aiofiles
import Stemmer  # type: ignore[import-not-found]
import structlog

from kodit.domain.services.bm25_service import BM25Repository
from kodit.domain.value_objects import (
    DeleteRequest,
    IndexRequest,
    SearchRequest,
    SearchResult,
)

if TYPE_CHECKING:
    import bm25s  # type: ignore[import-untyped]
    from bm25s.tokenization import Tokenized  # type: ignore[import-untyped]

SNIPPET_IDS_FILE = "snippet_ids.jsonl"


class LocalBM25Repository(BM25Repository):
    """Local BM25 repository implementation."""

    def __init__(self, data_dir: Path) -> None:
        """Initialize the local BM25 repository.

        Args:
            data_dir: Directory to store BM25 index files

        """
        self.log = structlog.get_logger(__name__)
        self.index_path = data_dir / "bm25s_index"
        self.snippet_ids: list[str] = []
        self.stemmer = Stemmer.Stemmer("english")
        self.__retriever: bm25s.BM25 | None = None

    def _retriever(self) -> bm25s.BM25:
        """Get the BM25 retriever."""
        if self.__retriever is None:
            import bm25s

            try:
                self.log.debug("Loading BM25 index")
                self.__retriever = bm25s.BM25.load(self.index_path, mmap=True)
                with Path(self.index_path / SNIPPET_IDS_FILE).open() as f:
                    self.snippet_ids = json.load(f)
            except FileNotFoundError:
                self.log.debug("BM25 index not found, creating new index")
                self.__retriever = bm25s.BM25()
        return self.__retriever

    def _tokenize(self, corpus: list[str]) -> list[list[str]] | Tokenized:
        """Tokenize text corpus."""
        from bm25s import tokenize

        return tokenize(
            corpus,
            stopwords="en",
            stemmer=self.stemmer,
            return_ids=False,
            show_progress=True,
            lower=True,
        )

    async def index_documents(self, request: IndexRequest) -> None:
        """Index documents for BM25 search."""
        self.log.debug("Indexing corpus")
        if not request.documents:
            self.log.warning("Corpus is empty, skipping bm25 index")
            return

        if not self.snippet_ids and (self.index_path / SNIPPET_IDS_FILE).exists():
            async with aiofiles.open(self.index_path / SNIPPET_IDS_FILE) as f:
                self.snippet_ids = json.loads(await f.read())

        # Filter out documents that have already been indexed
        new_documents = [
            doc for doc in request.documents if doc.snippet_id not in self.snippet_ids
        ]
        if not new_documents:
            self.log.info("No new documents to index")
            return

        vocab = self._tokenize([doc.text for doc in new_documents])
        self._retriever().index(vocab, show_progress=False)
        self._retriever().save(self.index_path)
        # Replace snippet_ids instead of appending, since the BM25 index is rebuilt
        self.snippet_ids = [doc.snippet_id for doc in new_documents]
        async with aiofiles.open(self.index_path / SNIPPET_IDS_FILE, "w") as f:
            await f.write(json.dumps(self.snippet_ids))

    async def search(self, request: SearchRequest) -> list[SearchResult]:
        """Search documents using BM25."""
        if request.top_k == 0:
            self.log.warning("Top k is 0, returning empty list")
            return []

        # Check that the index has data
        if not hasattr(self._retriever(), "scores"):
            return []

        # Get the number of documents in the index
        num_docs = self._retriever().scores["num_docs"]
        if num_docs == 0:
            return []

        # Adjust top_k to not exceed corpus size
        top_k = min(request.top_k, num_docs)
        self.log.debug(
            "Retrieving from index",
            query=request.query,
            top_k=top_k,
        )

        query_tokens = self._tokenize([request.query])

        self.log.debug("Query tokens", query_tokens=query_tokens)

        results, scores = self._retriever().retrieve(
            query_tokens=query_tokens,
            corpus=self.snippet_ids,
            k=top_k,
        )
        self.log.debug("Raw results", results=results, scores=scores)

        # Filter results by snippet_ids if provided
        filtered_results = []
        for result, score in zip(results[0], scores[0], strict=True):
            snippet_id = result
            if score > 0.0 and (
                request.snippet_ids is None or snippet_id in request.snippet_ids
            ):
                filtered_results.append(
                    SearchResult(snippet_id=snippet_id, score=float(score))
                )

        return filtered_results

    async def delete_documents(self, request: DeleteRequest) -> None:
        """Delete documents from the index."""
        # request parameter is unused as deletion is not supported
        # ruff: noqa: ARG002
        self.log.warning("Deletion not supported for local BM25 index")



"""Code elements extracted from parsed files."""

from dataclasses import dataclass
from pathlib import Path

from tree_sitter import Node, Tree

from kodit.domain.entities.git import GitFile


@dataclass
class ParsedFile:
    """Result of parsing a single file with tree-sitter."""

    path: Path
    git_file: GitFile
    tree: Tree
    source_code: bytes


@dataclass
class FunctionDefinition:
    """Information about a function or method definition."""

    file: Path
    node: Node
    span: tuple[int, int]
    qualified_name: str
    simple_name: str
    is_public: bool
    is_method: bool
    docstring: str | None
    parameters: list[str]
    return_type: str | None


@dataclass
class ClassDefinition:
    """Information about a class definition."""

    file: Path
    node: Node
    span: tuple[int, int]
    qualified_name: str
    simple_name: str
    is_public: bool
    docstring: str | None
    methods: list[FunctionDefinition]
    base_classes: list[str]
    constructor_params: list[str]


@dataclass
class TypeDefinition:
    """Information about a type definition (enum, interface, type alias)."""

    file: Path
    node: Node
    span: tuple[int, int]
    qualified_name: str
    simple_name: str
    is_public: bool
    docstring: str | None
    kind: str
    constructor_params: list[str]


@dataclass
class ModuleDefinition:
    """All definitions in a module, grouped by language conventions."""

    module_path: str
    files: list[ParsedFile]
    functions: list[FunctionDefinition]
    classes: list[ClassDefinition]
    types: list[TypeDefinition]
    constants: list[tuple[str, Node]]
    module_docstring: str | None


@dataclass
class FunctionInfo:
    """Information about a function definition."""

    file: Path
    node: Node
    span: tuple[int, int]
    qualified_name: str



"""Enrichment association repository."""

from collections.abc import Callable

import structlog
from sqlalchemy.ext.asyncio import AsyncSession

from kodit.domain.enrichments.enrichment import (
    EnrichmentAssociation,
)
from kodit.domain.protocols import EnrichmentAssociationRepository
from kodit.infrastructure.sqlalchemy import entities as db_entities
from kodit.infrastructure.sqlalchemy.repository import SqlAlchemyRepository


def create_enrichment_association_repository(
    session_factory: Callable[[], AsyncSession],
) -> EnrichmentAssociationRepository:
    """Create a enrichment association repository."""
    return SQLAlchemyEnrichmentAssociationRepository(session_factory=session_factory)


class SQLAlchemyEnrichmentAssociationRepository(
    SqlAlchemyRepository[EnrichmentAssociation, db_entities.EnrichmentAssociation],
    EnrichmentAssociationRepository,
):
    """Repository for managing enrichment associations."""

    def __init__(self, session_factory: Callable[[], AsyncSession]) -> None:
        """Initialize the repository."""
        super().__init__(session_factory=session_factory)
        self._log = structlog.get_logger(__name__)

    def _get_id(self, entity: EnrichmentAssociation) -> int | None:
        """Get the ID of an enrichment association."""
        return entity.id

    @property
    def db_entity_type(self) -> type[db_entities.EnrichmentAssociation]:
        """The SQLAlchemy model type."""
        return db_entities.EnrichmentAssociation

    @staticmethod
    def to_domain(
        db_entity: db_entities.EnrichmentAssociation,
    ) -> EnrichmentAssociation:
        """Map database entity to domain entity."""
        return EnrichmentAssociation(
            enrichment_id=db_entity.enrichment_id,
            entity_type=db_entity.entity_type,
            entity_id=db_entity.entity_id,
            id=db_entity.id,
        )

    @staticmethod
    def to_db(
        domain_entity: EnrichmentAssociation,
    ) -> db_entities.EnrichmentAssociation:
        """Map domain entity to database entity."""
        from datetime import UTC, datetime

        now = datetime.now(UTC)
        db_entity = db_entities.EnrichmentAssociation(
            enrichment_id=domain_entity.enrichment_id,
            entity_type=domain_entity.entity_type,
            entity_id=domain_entity.entity_id,
        )
        if domain_entity.id is not None:
            db_entity.id = domain_entity.id
        # Always set timestamps since domain entity doesn't track them
        db_entity.created_at = now
        db_entity.updated_at = now
        return db_entity



"""Base class and language-specific analyzers."""

from abc import ABC, abstractmethod
from pathlib import Path

from tree_sitter import Node

from kodit.infrastructure.slicing.code_elements import (
    ClassDefinition,
    FunctionDefinition,
    ParsedFile,
    TypeDefinition,
)
from kodit.infrastructure.slicing.language_config import (
    LanguageMetadata,
    LanguageNodeTypes,
)


class LanguageAnalyzer(ABC):
    """Base class for language-specific analysis."""

    @abstractmethod
    def node_types(self) -> LanguageNodeTypes:
        """Node types for this language."""

    @abstractmethod
    def metadata(self) -> LanguageMetadata:
        """Metadata for this language."""

    @abstractmethod
    def extract_function_name(self, node: Node) -> str | None:
        """Extract function name from a function definition node."""

    @abstractmethod
    def is_public(self, node: Node, name: str) -> bool:
        """Determine if a definition is public."""

    @abstractmethod
    def extract_docstring(self, node: Node) -> str | None:
        """Extract documentation comment for a definition."""

    @abstractmethod
    def extract_module_path(self, parsed: ParsedFile) -> str:
        """Extract module/package path based on language conventions."""

    def is_method(self, node: Node) -> bool:  # noqa: ARG002
        """Check if a function node represents a method."""
        return False

    def extract_module_docstring(
        self,
        module_files: list[ParsedFile],  # noqa: ARG002
    ) -> str | None:
        """Extract module-level documentation."""
        return None

    def extract_classes(
        self,
        parsed: ParsedFile,  # noqa: ARG002
        *,
        include_private: bool,  # noqa: ARG002
    ) -> list[ClassDefinition]:
        """Extract class definitions with their methods."""
        return []

    def extract_types(
        self,
        parsed: ParsedFile,  # noqa: ARG002
        *,
        include_private: bool,  # noqa: ARG002
    ) -> list[TypeDefinition]:
        """Extract type definitions."""
        return []

    def extract_constructor_params(self, node: Node) -> list[str]:  # noqa: ARG002
        """Extract constructor parameters from a class or type node."""
        return []


class PythonAnalyzer(LanguageAnalyzer):
    """Python-specific analyzer."""

    def node_types(self) -> LanguageNodeTypes:
        """Node types for Python."""
        return LanguageNodeTypes(
            function_nodes=("function_definition",),
            method_nodes=(),
            call_node="call",
            import_nodes=("import_statement", "import_from_statement"),
        )

    def metadata(self) -> LanguageMetadata:
        """Metadata for Python."""
        return LanguageMetadata(
            extension=".py",
            name_field=None,
            tree_sitter_name="python",
        )

    def extract_function_name(self, node: Node) -> str | None:
        """Extract function name from a function definition node."""
        for child in node.children:
            if child.type == "identifier" and child.text is not None:
                return child.text.decode("utf-8")
        return None

    def is_public(self, node: Node, name: str) -> bool:  # noqa: ARG002
        """Determine if a definition is public."""
        return not name.startswith("_")

    def extract_docstring(self, node: Node) -> str | None:  # noqa: C901
        """Extract Python docstring."""
        body_node = None

        if node.type in {"function_definition", "class_definition"}:
            for child in node.children:
                if child.type == "block":
                    body_node = child
                    break
        elif node.type == "module":
            body_node = node

        if not body_node:
            return None

        for child in body_node.children:
            if child.type == "expression_statement":
                for expr_child in child.children:
                    if expr_child.type == "string" and expr_child.text:
                        docstring_bytes = expr_child.text
                        try:
                            docstring_text = docstring_bytes.decode("utf-8")
                            docstring_text = docstring_text.strip()
                            for quote in ['"""', "'''", '"', "'"]:
                                starts = docstring_text.startswith(quote)
                                ends = docstring_text.endswith(quote)
                                if starts and ends:
                                    quote_len = len(quote)
                                    docstring_text = docstring_text[
                                        quote_len:-quote_len
                                    ]
                                    break
                            return docstring_text.strip()
                        except UnicodeDecodeError:
                            return None
                break

        return None

    def extract_module_path(self, parsed: ParsedFile) -> str:
        """Extract Python module path from file path structure."""
        file_path = Path(parsed.git_file.path)
        clean_path = self._clean_path_for_module(file_path)

        if clean_path.name == "__init__.py":
            module_parts = list(clean_path.parts[:-1])
        else:
            module_parts = list(clean_path.parts[:-1])
            module_parts.append(clean_path.stem)

        module_path = ".".join(p for p in module_parts if p and p != ".")

        if not module_path and clean_path.name == "__init__.py":
            parent_dir = file_path.parent.name
            if parent_dir and parent_dir != ".":
                return parent_dir
            return ""

        return module_path if module_path else clean_path.stem

    def extract_module_docstring(self, module_files: list[ParsedFile]) -> str | None:
        """Extract Python module-level documentation."""
        for parsed in module_files:
            if parsed.path.name == "__init__.py":
                return self.extract_docstring(parsed.tree.root_node)
        if module_files:
            return self.extract_docstring(module_files[0].tree.root_node)
        return None

    def _clean_path_for_module(self, file_path: Path) -> Path:
        """Clean a file path to extract a reasonable module path."""
        parts = list(file_path.parts)

        if not file_path.is_absolute():
            return file_path

        common_roots = {"src", "lib", "pkg", "internal", "app"}
        for i, part in enumerate(parts):
            if part in common_roots:
                if i + 1 < len(parts):
                    return Path(*parts[i + 1 :])
                return file_path

        if len(parts) >= 2:
            return Path(*parts[-2:])

        return file_path

    def extract_classes(
        self, parsed: ParsedFile, *, include_private: bool
    ) -> list[ClassDefinition]:
        """Extract Python class definitions."""
        classes = []

        for node in self._walk_tree(parsed.tree.root_node):
            if node.type == "class_definition":
                class_name = None
                for child in node.children:
                    if child.type == "identifier" and child.text:
                        class_name = child.text.decode("utf-8")
                        break

                if not class_name:
                    continue

                is_public = self.is_public(node, class_name)
                if not include_private and not is_public:
                    continue

                docstring = self.extract_docstring(node)
                methods = self._extract_class_methods(node, parsed, include_private)
                base_classes = self._extract_base_classes(node)

                qualified_name = f"{parsed.path.stem}.{class_name}"
                span = (node.start_byte, node.end_byte)

                constructor_params = self.extract_constructor_params(node)

                classes.append(
                    ClassDefinition(
                        file=parsed.path,
                        node=node,
                        span=span,
                        qualified_name=qualified_name,
                        simple_name=class_name,
                        is_public=is_public,
                        docstring=docstring,
                        methods=methods,
                        base_classes=base_classes,
                        constructor_params=constructor_params,
                    )
                )

        return classes

    def _walk_tree(self, node: Node):  # noqa: ANN202
        """Walk the AST tree, yielding all nodes."""
        queue = [node]
        visited: set[int] = set()

        while queue:
            current = queue.pop(0)
            node_id = id(current)
            if node_id in visited:
                continue
            visited.add(node_id)

            yield current
            queue.extend(current.children)

    def _extract_class_methods(  # noqa: ANN202, C901
        self,
        class_node: Node,
        parsed: ParsedFile,
        include_private: bool,  # noqa: FBT001
    ):
        """Extract methods from a class definition."""
        methods = []

        for child in class_node.children:
            if child.type == "block":
                for block_child in child.children:
                    if block_child.type == "function_definition":
                        method_name = None
                        for func_child in block_child.children:
                            if func_child.type == "identifier" and func_child.text:
                                method_name = func_child.text.decode("utf-8")
                                break

                        if not method_name:
                            continue

                        is_public = self.is_public(block_child, method_name)
                        if not include_private and not is_public:
                            continue

                        docstring = self.extract_docstring(block_child)
                        parameters = self._extract_parameters_from_function(block_child)

                        class_name = None
                        for class_child in class_node.children:
                            if class_child.type == "identifier" and class_child.text:
                                class_name = class_child.text.decode("utf-8")
                                break

                        qualified_name = (
                            f"{parsed.path.stem}.{class_name}.{method_name}"
                        )
                        span = (block_child.start_byte, block_child.end_byte)

                        method = FunctionDefinition(
                            file=parsed.path,
                            node=block_child,
                            span=span,
                            qualified_name=qualified_name,
                            simple_name=method_name,
                            is_public=is_public,
                            is_method=True,
                            docstring=docstring,
                            parameters=parameters,
                            return_type=None,
                        )
                        methods.append(method)

        return methods

    def _extract_base_classes(self, class_node: Node) -> list[str]:
        """Extract base class names from a class definition."""
        base_classes: list[str] = []

        for child in class_node.children:
            if child.type == "argument_list":
                base_classes.extend(
                    arg_child.text.decode("utf-8")
                    for arg_child in child.children
                    if arg_child.type == "identifier" and arg_child.text
                )

        return base_classes

    def extract_constructor_params(self, node: Node) -> list[str]:
        """Extract constructor parameters from Python __init__ method."""
        for child in self._walk_tree(node):
            if child.type == "function_definition":
                func_name = self.extract_function_name(child)
                if func_name == "__init__":
                    return self._extract_parameters_from_function(child)
        return []

    def _extract_parameters_from_function(self, func_node: Node) -> list[str]:
        """Extract parameter names and types from a function definition."""
        params: list[str] = []
        for child in func_node.children:
            if child.type == "parameters":
                for param_child in child.children:
                    if param_child.type in {
                        "identifier",
                        "typed_parameter",
                        "typed_default_parameter",
                        "default_parameter",
                    }:
                        param_text = self._extract_param_text(param_child)
                        if param_text and param_text != "self":
                            params.append(param_text)
        return params

    def _extract_param_text(self, param_node: Node) -> str | None:
        """Extract parameter text including type annotation."""
        if param_node.text:
            try:
                return param_node.text.decode("utf-8")
            except UnicodeDecodeError:
                return None
        return None


class GoAnalyzer(LanguageAnalyzer):
    """Go-specific analyzer."""

    def node_types(self) -> LanguageNodeTypes:
        """Node types for Go."""
        return LanguageNodeTypes(
            function_nodes=("function_declaration",),
            method_nodes=("method_declaration",),
            call_node="call_expression",
            import_nodes=("import_declaration",),
        )

    def metadata(self) -> LanguageMetadata:
        """Metadata for Go."""
        return LanguageMetadata(
            extension=".go",
            name_field=None,
            tree_sitter_name="go",
        )

    def extract_function_name(self, node: Node) -> str | None:
        """Extract function name from a function definition node."""
        if node.type == "method_declaration":
            for child in node.children:
                if child.type == "field_identifier" and child.text is not None:
                    return child.text.decode("utf-8")
        else:
            for child in node.children:
                if child.type == "identifier" and child.text is not None:
                    return child.text.decode("utf-8")
        return None

    def is_public(self, node: Node, name: str) -> bool:  # noqa: ARG002
        """Determine if a definition is public."""
        return name[0].isupper() if name else False

    def is_method(self, node: Node) -> bool:
        """Check if a function node represents a method."""
        return node.type == "method_declaration"

    def extract_docstring(self, node: Node) -> str | None:
        """Extract comment before a Go function or method declaration."""
        parent = node.parent
        if not parent:
            return None

        func_index = None
        for i, child in enumerate(parent.children):
            if child == node:
                func_index = i
                break

        if func_index is None or func_index == 0:
            return None

        prev_sibling = parent.children[func_index - 1]
        if prev_sibling.type == "comment" and prev_sibling.text:
            comment_text = prev_sibling.text.decode("utf-8")
            return comment_text.lstrip("/").strip()

        return None

    def extract_module_path(self, parsed: ParsedFile) -> str:
        """Extract full Go package path using directory structure."""
        package_name = self._extract_package_name(parsed)

        file_path = Path(parsed.git_file.path)
        clean_path = self._clean_path_for_module(file_path)
        dir_path = clean_path.parent

        if str(dir_path) != ".":
            dir_str = str(dir_path).replace("\\", "/")
            if dir_str.endswith("/" + package_name) or dir_str == package_name:
                return dir_str
            return f"{dir_str}/{package_name}"
        return package_name

    def _extract_package_name(self, parsed: ParsedFile) -> str:
        """Extract Go package name from package declaration."""
        root = parsed.tree.root_node
        for child in root.children:
            if child.type == "package_clause":
                for package_child in child.children:
                    if (
                        package_child.type == "package_identifier"
                        and package_child.text
                    ):
                        return package_child.text.decode("utf-8")
        return parsed.path.stem

    def _clean_path_for_module(self, file_path: Path) -> Path:
        """Clean a file path to extract a reasonable module path."""
        parts = list(file_path.parts)

        if not file_path.is_absolute():
            return file_path

        common_roots = {"src", "lib", "pkg", "internal", "app"}
        for i, part in enumerate(parts):
            if part in common_roots:
                if i + 1 < len(parts):
                    return Path(*parts[i + 1 :])
                return file_path

        if len(parts) >= 2:
            return Path(*parts[-2:])

        return file_path

    def extract_types(
        self, parsed: ParsedFile, *, include_private: bool
    ) -> list[TypeDefinition]:
        """Extract Go type definitions."""
        types = []

        for node in self._walk_tree(parsed.tree.root_node):
            if node.type == "type_declaration":
                for child in node.children:
                    if child.type == "type_spec":
                        type_def = self._extract_type_from_spec(
                            child, parsed, include_private=include_private
                        )
                        if type_def:
                            types.append(type_def)

        return types

    def _walk_tree(self, node: Node):  # noqa: ANN202
        """Walk the AST tree, yielding all nodes."""
        queue = [node]
        visited: set[int] = set()

        while queue:
            current = queue.pop(0)
            node_id = id(current)
            if node_id in visited:
                continue
            visited.add(node_id)

            yield current
            queue.extend(current.children)

    def _extract_type_from_spec(
        self, type_spec_node: Node, parsed: ParsedFile, *, include_private: bool
    ) -> TypeDefinition | None:
        """Extract a single Go type definition from a type_spec node."""
        type_name = None
        type_kind = "type"

        for child in type_spec_node.children:
            if child.type == "type_identifier" and child.text:
                type_name = child.text.decode("utf-8")
            elif child.type == "struct_type":
                type_kind = "struct"
            elif child.type == "interface_type":
                type_kind = "interface"
            elif child.type in ["slice_type", "array_type", "pointer_type"]:
                type_kind = "alias"
            elif child.type == "map_type":
                type_kind = "map"

        if not type_name:
            return None

        is_public = self.is_public(type_spec_node, type_name)
        if not include_private and not is_public:
            return None

        parent_node = type_spec_node.parent
        if not parent_node:
            return None

        docstring = self._extract_type_comment(parent_node)

        qualified_name = f"{parsed.path.stem}.{type_name}"
        span = (parent_node.start_byte, parent_node.end_byte)
        constructor_params = self.extract_constructor_params(type_spec_node)

        return TypeDefinition(
            file=parsed.path,
            node=parent_node,
            span=span,
            qualified_name=qualified_name,
            simple_name=type_name,
            is_public=is_public,
            docstring=docstring,
            kind=type_kind,
            constructor_params=constructor_params,
        )

    def _extract_type_comment(self, type_decl_node: Node) -> str | None:
        """Extract comment before a Go type declaration."""
        parent = type_decl_node.parent
        if not parent:
            return None

        type_decl_index = None
        for i, child in enumerate(parent.children):
            if child == type_decl_node:
                type_decl_index = i
                break

        if type_decl_index is None or type_decl_index == 0:
            return None

        prev_sibling = parent.children[type_decl_index - 1]
        if prev_sibling.type == "comment" and prev_sibling.text:
            comment_text = prev_sibling.text.decode("utf-8")
            return comment_text.lstrip("/").strip()

        return None

    def extract_constructor_params(self, node: Node) -> list[str]:
        """Extract constructor parameters from Go struct fields."""
        params: list[str] = []

        for child in node.children:
            if child.type == "struct_type":
                for struct_child in child.children:
                    if struct_child.type == "field_declaration_list":
                        for field in struct_child.children:
                            if field.type == "field_declaration":
                                field_text = self._extract_go_field_signature(field)
                                if field_text:
                                    params.append(field_text)
        return params

    def _extract_go_field_signature(self, field_node: Node) -> str | None:
        """Extract field signature from a Go field declaration."""
        if field_node.text:
            try:
                return field_node.text.decode("utf-8").strip()
            except UnicodeDecodeError:
                return None
        return None


class JavaAnalyzer(LanguageAnalyzer):
    """Java-specific analyzer."""

    def node_types(self) -> LanguageNodeTypes:
        """Node types for Java."""
        return LanguageNodeTypes(
            function_nodes=("method_declaration",),
            method_nodes=(),
            call_node="method_invocation",
            import_nodes=("import_declaration",),
        )

    def metadata(self) -> LanguageMetadata:
        """Metadata for Java."""
        return LanguageMetadata(
            extension=".java",
            name_field=None,
            tree_sitter_name="java",
        )

    def extract_function_name(self, node: Node) -> str | None:
        """Extract function name from a function definition node."""
        for child in node.children:
            if child.type == "identifier" and child.text is not None:
                return child.text.decode("utf-8")
        return None

    def is_public(self, node: Node, name: str) -> bool:  # noqa: ARG002
        """Determine if a definition is public."""
        return True

    def extract_docstring(self, node: Node) -> str | None:  # noqa: ARG002
        """Extract documentation comment for a definition."""
        return None

    def extract_module_path(self, parsed: ParsedFile) -> str:
        """Extract Java package name from package declaration."""
        root = parsed.tree.root_node
        for child in root.children:
            if child.type == "package_declaration":
                for package_child in child.children:
                    if package_child.type == "scoped_identifier" and package_child.text:
                        return package_child.text.decode("utf-8")
                    if package_child.type == "identifier" and package_child.text:
                        return package_child.text.decode("utf-8")
        return parsed.path.stem


class CAnalyzer(LanguageAnalyzer):
    """C-specific analyzer."""

    def node_types(self) -> LanguageNodeTypes:
        """Node types for C."""
        return LanguageNodeTypes(
            function_nodes=("function_definition",),
            method_nodes=(),
            call_node="call_expression",
            import_nodes=("preproc_include",),
        )

    def metadata(self) -> LanguageMetadata:
        """Metadata for C."""
        return LanguageMetadata(
            extension=".c",
            name_field="declarator",
            tree_sitter_name="c",
        )

    def extract_function_name(self, node: Node) -> str | None:
        """Extract function name from a C function definition."""
        declarator = node.child_by_field_name("declarator")
        if not declarator:
            return None

        if declarator.type == "function_declarator":
            for child in declarator.children:
                if child.type == "identifier" and child.text is not None:
                    return child.text.decode("utf-8")
        elif declarator.type == "identifier" and declarator.text is not None:
            return declarator.text.decode("utf-8")
        return None

    def is_public(self, node: Node, name: str) -> bool:  # noqa: ARG002
        """Determine if a definition is public."""
        return True

    def extract_docstring(self, node: Node) -> str | None:  # noqa: ARG002
        """Extract documentation comment for a definition."""
        return None

    def extract_module_path(self, parsed: ParsedFile) -> str:
        """Extract module path based on file path."""
        file_path = Path(parsed.git_file.path)
        clean_path = self._clean_path_for_module(file_path)
        module_parts = list(clean_path.parts[:-1])
        module_parts.append(clean_path.stem)
        module_path = ".".join(p for p in module_parts if p and p != ".")
        return module_path if module_path else clean_path.stem

    def _clean_path_for_module(self, file_path: Path) -> Path:
        """Clean a file path to extract a reasonable module path."""
        parts = list(file_path.parts)

        if not file_path.is_absolute():
            return file_path

        common_roots = {"src", "lib", "pkg", "internal", "app"}
        for i, part in enumerate(parts):
            if part in common_roots:
                if i + 1 < len(parts):
                    return Path(*parts[i + 1 :])
                return file_path

        if len(parts) >= 2:
            return Path(*parts[-2:])

        return file_path


class CppAnalyzer(CAnalyzer):
    """C++-specific analyzer."""

    def node_types(self) -> LanguageNodeTypes:
        """Node types for C++."""
        return LanguageNodeTypes(
            function_nodes=("function_definition",),
            method_nodes=(),
            call_node="call_expression",
            import_nodes=("preproc_include", "using_declaration"),
        )

    def metadata(self) -> LanguageMetadata:
        """Metadata for C++."""
        return LanguageMetadata(
            extension=".cpp",
            name_field="declarator",
            tree_sitter_name="cpp",
        )


class RustAnalyzer(LanguageAnalyzer):
    """Rust-specific analyzer."""

    def node_types(self) -> LanguageNodeTypes:
        """Node types for Rust."""
        return LanguageNodeTypes(
            function_nodes=("function_item",),
            method_nodes=(),
            call_node="call_expression",
            import_nodes=("use_declaration", "extern_crate_declaration"),
        )

    def metadata(self) -> LanguageMetadata:
        """Metadata for Rust."""
        return LanguageMetadata(
            extension=".rs",
            name_field="name",
            tree_sitter_name="rust",
        )

    def extract_function_name(self, node: Node) -> str | None:
        """Extract function name from a Rust function definition."""
        name_node = node.child_by_field_name("name")
        if name_node and name_node.type == "identifier" and name_node.text is not None:
            return name_node.text.decode("utf-8")
        return None

    def is_public(self, node: Node, name: str) -> bool:  # noqa: ARG002
        """Determine if a definition is public."""
        return True

    def extract_docstring(self, node: Node) -> str | None:  # noqa: ARG002
        """Extract documentation comment for a definition."""
        return None

    def extract_module_path(self, parsed: ParsedFile) -> str:
        """Extract module path based on file path."""
        file_path = Path(parsed.git_file.path)
        clean_path = self._clean_path_for_module(file_path)
        module_parts = list(clean_path.parts[:-1])
        module_parts.append(clean_path.stem)
        module_path = ".".join(p for p in module_parts if p and p != ".")
        return module_path if module_path else clean_path.stem

    def _clean_path_for_module(self, file_path: Path) -> Path:
        """Clean a file path to extract a reasonable module path."""
        parts = list(file_path.parts)

        if not file_path.is_absolute():
            return file_path

        common_roots = {"src", "lib", "pkg", "internal", "app"}
        for i, part in enumerate(parts):
            if part in common_roots:
                if i + 1 < len(parts):
                    return Path(*parts[i + 1 :])
                return file_path

        if len(parts) >= 2:
            return Path(*parts[-2:])

        return file_path


class JavaScriptAnalyzer(LanguageAnalyzer):
    """JavaScript-specific analyzer."""

    def node_types(self) -> LanguageNodeTypes:
        """Node types for JavaScript."""
        return LanguageNodeTypes(
            function_nodes=(
                "function_declaration",
                "function_expression",
                "arrow_function",
            ),
            method_nodes=(),
            call_node="call_expression",
            import_nodes=("import_statement", "import_declaration"),
        )

    def metadata(self) -> LanguageMetadata:
        """Metadata for JavaScript."""
        return LanguageMetadata(
            extension=".js",
            name_field=None,
            tree_sitter_name="javascript",
        )

    def extract_function_name(self, node: Node) -> str | None:
        """Extract function name from a function definition node."""
        # For function declarations, the name is a direct child
        if node.type in ("function_declaration", "function_expression"):
            for child in node.children:
                if child.type == "identifier" and child.text is not None:
                    return child.text.decode("utf-8")
            return None

        # For arrow functions, check if it's a named binding (const foo = () => {})
        if node.type == "arrow_function":
            return self._get_arrow_function_name(node)

        return None

    def _get_arrow_function_name(self, node: Node) -> str | None:
        """Get the name of an arrow function from its variable binding."""
        # Arrow function should be inside a variable_declarator
        # Pattern: const foo = () => {} or let foo = () => {}
        parent = node.parent
        if parent is None:
            return None

        if parent.type == "variable_declarator":
            # Find the identifier sibling
            for child in parent.children:
                if child.type == "identifier" and child.text is not None:
                    return child.text.decode("utf-8")

        # Not a named arrow function (e.g., inline callback like .filter(x => x))
        return None

    def is_public(self, node: Node, name: str) -> bool:  # noqa: ARG002
        """Determine if a definition is public."""
        return True

    def extract_docstring(self, node: Node) -> str | None:  # noqa: ARG002
        """Extract documentation comment for a definition."""
        return None

    def _walk_tree(self, node: Node):  # noqa: ANN202
        """Walk the AST tree, yielding all nodes."""
        queue = [node]
        visited: set[int] = set()

        while queue:
            current = queue.pop(0)
            node_id = id(current)
            if node_id in visited:
                continue
            visited.add(node_id)

            yield current
            queue.extend(current.children)

    def extract_module_path(self, parsed: ParsedFile) -> str:
        """Extract module path based on file path."""
        file_path = Path(parsed.git_file.path)
        clean_path = self._clean_path_for_module(file_path)
        module_parts = list(clean_path.parts[:-1])
        module_parts.append(clean_path.stem)
        module_path = ".".join(p for p in module_parts if p and p != ".")
        return module_path if module_path else clean_path.stem

    def _clean_path_for_module(self, file_path: Path) -> Path:
        """Clean a file path to extract a reasonable module path."""
        parts = list(file_path.parts)

        if not file_path.is_absolute():
            return file_path

        common_roots = {"src", "lib", "pkg", "internal", "app"}
        for i, part in enumerate(parts):
            if part in common_roots:
                if i + 1 < len(parts):
                    return Path(*parts[i + 1 :])
                return file_path

        if len(parts) >= 2:
            return Path(*parts[-2:])

        return file_path


class TypeScriptAnalyzer(JavaScriptAnalyzer):
    """TypeScript-specific analyzer."""

    def metadata(self) -> LanguageMetadata:
        """Metadata for TypeScript."""
        return LanguageMetadata(
            extension=".ts",
            name_field=None,
            tree_sitter_name="typescript",
        )

    def extract_types(
        self, parsed: ParsedFile, *, include_private: bool  # noqa: ARG002
    ) -> list[TypeDefinition]:
        """Extract TypeScript interface and type alias definitions."""
        types = []

        for node in self._walk_tree(parsed.tree.root_node):
            if node.type in ("interface_declaration", "type_alias_declaration"):
                type_name = self._extract_type_name(node)
                if not type_name:
                    continue

                module_name = parsed.path.stem
                qualified_name = f"{module_name}.{type_name}"
                kind = "interface" if node.type == "interface_declaration" else "type"

                types.append(
                    TypeDefinition(
                        file=parsed.path,
                        node=node,
                        span=(node.start_byte, node.end_byte),
                        qualified_name=qualified_name,
                        simple_name=type_name,
                        is_public=True,
                        docstring=None,
                        kind=kind,
                        constructor_params=[],
                    )
                )

        return types

    def _extract_type_name(self, node: Node) -> str | None:
        """Extract the name from an interface or type alias declaration."""
        for child in node.children:
            if child.type == "type_identifier" and child.text is not None:
                return child.text.decode("utf-8")
        return None

    def extract_type_references(self, node: Node) -> set[str]:
        """Extract type names referenced in a function's signature."""
        type_refs: set[str] = set()

        # Built-in types to exclude
        builtins = {
            "string",
            "number",
            "boolean",
            "void",
            "null",
            "undefined",
            "any",
            "never",
            "unknown",
            "object",
            "symbol",
            "bigint",
        }

        for child in self._walk_tree(node):
            if child.type == "type_identifier" and child.text is not None:
                type_name = child.text.decode("utf-8")
                if type_name.lower() not in builtins:
                    type_refs.add(type_name)

        return type_refs


class TsxAnalyzer(TypeScriptAnalyzer):
    """TSX-specific analyzer (TypeScript with JSX)."""

    def metadata(self) -> LanguageMetadata:
        """Metadata for TSX."""
        return LanguageMetadata(
            extension=".tsx",
            name_field=None,
            tree_sitter_name="tsx",
        )


class CSharpAnalyzer(LanguageAnalyzer):
    """C#-specific analyzer."""

    def node_types(self) -> LanguageNodeTypes:
        """Node types for C#."""
        return LanguageNodeTypes(
            function_nodes=("method_declaration",),
            method_nodes=("constructor_declaration",),
            call_node="invocation_expression",
            import_nodes=("using_directive",),
        )

    def metadata(self) -> LanguageMetadata:
        """Metadata for C#."""
        return LanguageMetadata(
            extension=".cs",
            name_field=None,
            tree_sitter_name="csharp",
        )

    def extract_function_name(self, node: Node) -> str | None:
        """Extract function name from a function definition node."""
        for child in node.children:
            if child.type == "identifier" and child.text is not None:
                return child.text.decode("utf-8")
        return None

    def is_public(self, node: Node, name: str) -> bool:  # noqa: ARG002
        """Determine if a definition is public."""
        return True

    def extract_docstring(self, node: Node) -> str | None:  # noqa: ARG002
        """Extract documentation comment for a definition."""
        return None

    def extract_module_path(self, parsed: ParsedFile) -> str:
        """Extract module path based on file path."""
        file_path = Path(parsed.git_file.path)
        clean_path = self._clean_path_for_module(file_path)
        module_parts = list(clean_path.parts[:-1])
        module_parts.append(clean_path.stem)
        module_path = ".".join(p for p in module_parts if p and p != ".")
        return module_path if module_path else clean_path.stem

    def _clean_path_for_module(self, file_path: Path) -> Path:
        """Clean a file path to extract a reasonable module path."""
        parts = list(file_path.parts)

        if not file_path.is_absolute():
            return file_path

        common_roots = {"src", "lib", "pkg", "internal", "app"}
        for i, part in enumerate(parts):
            if part in common_roots:
                if i + 1 < len(parts):
                    return Path(*parts[i + 1 :])
                return file_path

        if len(parts) >= 2:
            return Path(*parts[-2:])

        return file_path


def language_analyzer_factory(language: str) -> LanguageAnalyzer:
    """Create language-specific analyzer."""
    language_lower = language.lower()

    analyzers = {
        "python": PythonAnalyzer(),
        "go": GoAnalyzer(),
        "java": JavaAnalyzer(),
        "c": CAnalyzer(),
        "cpp": CppAnalyzer(),
        "c++": CppAnalyzer(),
        "rust": RustAnalyzer(),
        "javascript": JavaScriptAnalyzer(),
        "js": JavaScriptAnalyzer(),
        "typescript": TypeScriptAnalyzer(),
        "ts": TypeScriptAnalyzer(),
        "tsx": TsxAnalyzer(),
        "csharp": CSharpAnalyzer(),
        "c#": CSharpAnalyzer(),
        "cs": CSharpAnalyzer(),
    }

    analyzer = analyzers.get(language_lower)
    if not analyzer:
        raise ValueError(f"Unsupported language: {language}")

    return analyzer



"""Example file discovery from repositories."""

from pathlib import Path
from typing import ClassVar


class ExampleDiscovery:
    """Discovers example files in repositories."""

    EXAMPLE_DIRECTORIES: ClassVar[set[str]] = {
        "examples",
        "example",
        "samples",
        "sample",
        "demos",
        "demo",
        "tutorials",
        "tutorial",
    }

    DOCUMENTATION_EXTENSIONS: ClassVar[set[str]] = {
        ".md",
        ".markdown",
        ".rst",
        ".adoc",
        ".asciidoc",
    }

    def is_example_directory_file(self, file_path: str) -> bool:
        """Check if file is in an example directory."""
        path = Path(file_path)
        parts = [p.lower() for p in path.parts]
        return any(part in self.EXAMPLE_DIRECTORIES for part in parts)

    def is_documentation_file(self, file_path: str) -> bool:
        """Check if file is a documentation file."""
        return Path(file_path).suffix.lower() in self.DOCUMENTATION_EXTENSIONS

    def is_example_candidate(self, file_path: str) -> bool:
        """Check if file should be processed for examples."""
        return self.is_example_directory_file(file_path) or self.is_documentation_file(
            file_path
        )



"""Tests for LanguageMapping value object."""

from datetime import datetime

import pytest

from kodit.domain.value_objects import LanguageMapping, SnippetSearchFilters


class TestLanguageMapping:
    """Test cases for LanguageMapping value object."""

    def test_get_extensions_for_language_python(self) -> None:
        """Test getting extensions for Python."""
        extensions = LanguageMapping.get_extensions_for_language("python")
        assert extensions == ["py", "pyw", "pyx", "pxd"]

    def test_get_extensions_for_language_javascript(self) -> None:
        """Test getting extensions for JavaScript."""
        extensions = LanguageMapping.get_extensions_for_language("javascript")
        assert extensions == ["js", "jsx", "mjs"]

    def test_get_extensions_for_language_case_insensitive(self) -> None:
        """Test that language names are case insensitive."""
        extensions = LanguageMapping.get_extensions_for_language("PYTHON")
        assert extensions == ["py", "pyw", "pyx", "pxd"]

    def test_get_extensions_for_unsupported_language(self) -> None:
        """Test that unsupported languages raise ValueError."""
        with pytest.raises(ValueError, match="Unsupported language: unsupported"):
            LanguageMapping.get_extensions_for_language("unsupported")

    def test_get_language_for_extension_py(self) -> None:
        """Test getting language for .py extension."""
        language = LanguageMapping.get_language_for_extension("py")
        assert language == "python"

    def test_get_language_for_extension_with_dot(self) -> None:
        """Test getting language for extension with leading dot."""
        language = LanguageMapping.get_language_for_extension(".py")
        assert language == "python"

    def test_get_language_for_extension_case_insensitive(self) -> None:
        """Test that extensions are case insensitive."""
        language = LanguageMapping.get_language_for_extension("PY")
        assert language == "python"

    def test_get_language_for_unsupported_extension(self) -> None:
        """Test that unsupported extensions raise ValueError."""
        with pytest.raises(ValueError, match="Unsupported file extension: unsupported"):
            LanguageMapping.get_language_for_extension("unsupported")

    def test_get_extension_to_language_map(self) -> None:
        """Test getting the extension to language mapping."""
        extension_map = LanguageMapping.get_extension_to_language_map()

        # Check a few key mappings
        assert extension_map["py"] == "python"
        assert extension_map["js"] == "javascript"
        assert extension_map["go"] == "go"

        # Check that all extensions are included
        assert "py" in extension_map
        assert "js" in extension_map
        assert "ts" in extension_map

    def test_get_supported_languages(self) -> None:
        """Test getting list of supported languages."""
        languages = LanguageMapping.get_supported_languages()

        # Check that key languages are included
        assert "python" in languages
        assert "javascript" in languages
        assert "go" in languages
        assert "rust" in languages

    def test_get_supported_extensions(self) -> None:
        """Test getting list of supported extensions."""
        extensions = LanguageMapping.get_supported_extensions()

        # Check that key extensions are included
        assert "py" in extensions
        assert "js" in extensions
        assert "go" in extensions
        assert "rs" in extensions

    def test_is_supported_language(self) -> None:
        """Test checking if a language is supported."""
        assert LanguageMapping.is_supported_language("python") is True
        assert LanguageMapping.is_supported_language("PYTHON") is True
        assert LanguageMapping.is_supported_language("unsupported") is False

    def test_is_supported_extension(self) -> None:
        """Test checking if an extension is supported."""
        assert LanguageMapping.is_supported_extension("py") is True
        assert LanguageMapping.is_supported_extension(".py") is True
        assert LanguageMapping.is_supported_extension("PY") is True
        assert LanguageMapping.is_supported_extension("unsupported") is False

    def test_bidirectional_mapping_consistency(self) -> None:
        """Test that bidirectional mapping is consistent."""
        # Test that extension -> language -> extension gives the same result
        for language in LanguageMapping.get_supported_languages():
            extensions = LanguageMapping.get_extensions_for_language(language)
            for extension in extensions:
                detected_language = LanguageMapping.get_language_for_extension(
                    extension
                )
                assert detected_language == language

    def test_extension_uniqueness(self) -> None:
        """Test that each extension maps to only one language."""
        extension_map = LanguageMapping.get_extension_to_language_map()

        # Check that there are no duplicate extensions
        assert len(extension_map) == len(set(extension_map.keys()))

    def test_get_extensions_with_fallback_supported_language(self) -> None:
        """Test fallback method returns extensions for supported language."""
        extensions = LanguageMapping.get_extensions_with_fallback("python")
        assert extensions == ["py", "pyw", "pyx", "pxd"]

    def test_get_extensions_with_fallback_unsupported_language(self) -> None:
        """Test fallback method returns [language.lower()] for unsupported language."""
        extensions = LanguageMapping.get_extensions_with_fallback("foobar")
        assert extensions == ["foobar"]


class TestSnippetSearchFilters:
    """Test cases for SnippetSearchFilters value object."""

    def test_from_cli_params_no_filters(self) -> None:
        """Test that None is returned when no filters are provided."""
        filters = SnippetSearchFilters.from_cli_params()
        assert filters is None

    def test_from_cli_params_with_language(self) -> None:
        """Test creating filters with language parameter."""
        filters = SnippetSearchFilters.from_cli_params(language="python")
        assert filters is not None
        assert filters.language == "python"
        assert filters.author is None
        assert filters.created_after is None
        assert filters.created_before is None
        assert filters.source_repo is None

    def test_from_cli_params_with_all_filters(self) -> None:
        """Test creating filters with all parameters."""
        filters = SnippetSearchFilters.from_cli_params(
            language="python",
            author="John Doe",
            created_after="2023-01-01",
            created_before="2023-12-31",
            source_repo="github.com/example/repo",
        )
        assert filters is not None
        assert filters.language == "python"
        assert filters.author == "John Doe"
        assert filters.created_after == datetime(2023, 1, 1)  # noqa: DTZ001
        assert filters.created_before == datetime(2023, 12, 31)  # noqa: DTZ001
        assert filters.source_repo == "github.com/example/repo"

    def test_from_cli_params_invalid_date_format(self) -> None:
        """Test that invalid date formats raise ValueError."""
        with pytest.raises(ValueError, match="Invalid date format for created_after"):
            SnippetSearchFilters.from_cli_params(created_after="invalid-date")

        with pytest.raises(ValueError, match="Invalid date format for created_before"):
            SnippetSearchFilters.from_cli_params(created_before="invalid-date")

    def test_from_cli_params_partial_filters(self) -> None:
        """Test creating filters with only some parameters."""
        filters = SnippetSearchFilters.from_cli_params(
            language="go", author="Jane Smith"
        )
        assert filters is not None
        assert filters.language == "go"
        assert filters.author == "Jane Smith"
        assert filters.created_after is None
        assert filters.created_before is None
        assert filters.source_repo is None



"""Enrichment JSON-API schemas."""

from datetime import datetime

from pydantic import BaseModel, Field


class EnrichmentAttributes(BaseModel):
    """Enrichment attributes following JSON-API spec."""

    type: str
    subtype: str | None
    content: str
    created_at: datetime | None
    updated_at: datetime | None


class EnrichmentAssociationData(BaseModel):
    """Enrichment association data for JSON-API spec."""

    id: str
    type: str


class EnrichmentLinks(BaseModel):
    """Links following JSON-API spec."""

    self_link: str = Field(..., alias="self")

    model_config = {"populate_by_name": True}


class RelationshipData(BaseModel):
    """Data for a single relationship."""

    type: str
    id: str


class Relationship(BaseModel):
    """A JSON:API relationship."""

    links: EnrichmentLinks | None = None
    data: RelationshipData | list[RelationshipData] | None = None


class EnrichmentRelationships(BaseModel):
    """Enrichment relationships for JSON-API spec."""

    associations: list[EnrichmentAssociationData] | None = None
    commit: Relationship | None = None


class EnrichmentData(BaseModel):
    """Enrichment data following JSON-API spec."""

    type: str = "enrichment"
    id: str
    attributes: EnrichmentAttributes
    relationships: EnrichmentRelationships | None = None
    links: EnrichmentLinks | None = None


class EnrichmentListResponse(BaseModel):
    """Enrichment list response following JSON-API spec."""

    data: list[EnrichmentData]


class EnrichmentResponse(BaseModel):
    """Single enrichment response following JSON-API spec."""

    data: EnrichmentData


class EnrichmentUpdateAttributes(BaseModel):
    """Attributes for updating an enrichment."""

    content: str


class EnrichmentUpdateData(BaseModel):
    """Data for updating an enrichment."""

    type: str = "enrichment"
    attributes: EnrichmentUpdateAttributes


class EnrichmentUpdateRequest(BaseModel):
    """Request to update an enrichment."""

    data: EnrichmentUpdateData



"""Repository protocol interfaces for the domain layer."""

from abc import ABC, abstractmethod
from datetime import datetime
from pathlib import Path
from typing import Any, Protocol, TypeVar

from kodit.domain.enrichments.enrichment import EnrichmentAssociation, EnrichmentV2
from kodit.domain.entities import (
    Task,
    TaskStatus,
)
from kodit.domain.entities.git import (
    GitBranch,
    GitCommit,
    GitFile,
    GitRepo,
    GitTag,
    SnippetV2,
)
from kodit.domain.value_objects import (
    FusionRequest,
    FusionResult,
    MultiSearchRequest,
)
from kodit.infrastructure.sqlalchemy.query import Query


class GitAdapter(Protocol):
    """Protocol for Git repository operations."""

    async def clone_repository(self, remote_uri: str, local_path: Path) -> None:
        """Clone a repository to local path."""
        ...

    async def checkout_commit(self, local_path: Path, commit_sha: str) -> None:
        """Checkout a specific commit."""
        ...

    async def checkout_branch(self, local_path: Path, branch_name: str) -> None:
        """Checkout a specific branch."""
        ...

    async def fetch_repository(self, local_path: Path) -> None:
        """Fetch latest changes for existing repository."""
        ...

    async def pull_repository(self, local_path: Path) -> None:
        """Pull latest changes for existing repository."""
        ...

    async def get_all_branches(self, local_path: Path) -> list[dict[str, Any]]:
        """Get all branches in repository."""
        ...

    async def get_branch_commits(
        self, local_path: Path, branch_name: str
    ) -> list[dict[str, Any]]:
        """Get commit history for a specific branch."""
        ...

    async def get_all_commits_bulk(
        self, local_path: Path, since_date: datetime | None = None
    ) -> dict[str, dict[str, Any]]:
        """Get all commits from all branches in bulk for efficiency."""
        ...

    async def get_branch_commit_shas(
        self, local_path: Path, branch_name: str
    ) -> list[str]:
        """Get only commit SHAs for a branch."""
        ...

    async def get_all_branch_head_shas(
        self, local_path: Path, branch_names: list[str]
    ) -> dict[str, str]:
        """Get head commit SHAs for all branches in one operation."""
        ...

    async def get_commit_files(
        self, local_path: Path, commit_sha: str
    ) -> list[dict[str, Any]]:
        """Get all files in a specific commit from the git tree."""
        ...

    async def repository_exists(self, local_path: Path) -> bool:
        """Check if repository exists at local path."""
        ...

    async def get_commit_details(
        self, local_path: Path, commit_sha: str
    ) -> dict[str, Any]:
        """Get detailed information about a specific commit."""
        ...

    async def ensure_repository(self, remote_uri: str, local_path: Path) -> None:
        """Clone repository if it doesn't exist, otherwise pull latest changes."""
        ...

    async def get_file_content(
        self, local_path: Path, commit_sha: str, file_path: str
    ) -> bytes:
        """Get file content at specific commit."""
        ...

    async def get_default_branch(self, local_path: Path) -> str:
        """Get the default branch name with fallback strategies."""
        ...

    async def get_latest_commit_sha(
        self, local_path: Path, branch_name: str = "HEAD"
    ) -> str:
        """Get the latest commit SHA for a branch."""
        ...

    async def get_all_tags(self, local_path: Path) -> list[dict[str, Any]]:
        """Get all tags in repository."""
        ...

    async def get_commit_diff(self, local_path: Path, commit_sha: str) -> str:
        """Get the diff for a specific commit."""
        ...

T = TypeVar("T")


class Repository[T](Protocol):
    """Abstract base classes for repositories."""

    async def get(self, entity_id: Any) -> T:
        """Get entity by primary key."""
        ...

    async def get_or_create(
        self, entity: T, unique_field: str
    ) -> tuple[T, bool]:
        """Get existing entity or create new one atomically.

        Handles race conditions by catching unique constraint violations.
        Returns (entity, created) where created is True if new.
        """
        ...

    async def find(self, query: Query) -> list[T]:
        """Find all entities matching query."""
        ...

    async def save(self, entity: T) -> T:
        """Save entity (create new or update existing)."""
        ...

    async def save_bulk(
        self, entities: list[T], *, skip_existence_check: bool = False
    ) -> list[T]:
        """Save multiple entities in bulk (create new or update existing)."""
        ...

    async def exists(self, entity_id: Any) -> bool:
        """Check if entity exists by primary key."""
        ...

    async def delete(self, entity: T) -> None:
        """Remove entity."""
        ...

    async def delete_by_query(self, query: Query) -> None:
        """Remove entities by query."""
        ...

    async def count(self, query: Query) -> int:
        """Count the number of entities matching query."""
        ...


class TaskRepository(Repository[Task], Protocol):
    """Repository interface for Task entities."""

    async def next(self) -> Task | None:
        """Take a task for processing."""


class ReportingModule(Protocol):
    """Reporting module."""

    async def on_change(self, progress: TaskStatus) -> None:
        """On step changed."""
        ...


class TaskStatusRepository(Repository[TaskStatus]):
    """Repository interface for persisting progress state only."""

    @abstractmethod
    async def load_with_hierarchy(
        self, trackable_type: str, trackable_id: int
    ) -> list[TaskStatus]:
        """Load progress states with IDs and parent IDs from database."""

    @abstractmethod
    async def delete(self, entity: TaskStatus) -> None:
        """Delete a progress state."""


class GitCommitRepository(Repository[GitCommit]):
    """Repository for Git commits."""


class GitFileRepository(Repository[GitFile]):
    """Repository for Git files."""

    @abstractmethod
    async def delete_by_commit_sha(self, commit_sha: str) -> None:
        """Delete all files for a commit."""


class GitBranchRepository(Repository[GitBranch]):
    """Repository for Git branches."""

    @abstractmethod
    async def get_by_name(self, branch_name: str, repo_id: int) -> GitBranch:
        """Get a branch by name and repository ID."""

    @abstractmethod
    async def get_by_repo_id(self, repo_id: int) -> list[GitBranch]:
        """Get all branches for a repository."""

    @abstractmethod
    async def delete_by_repo_id(self, repo_id: int) -> None:
        """Delete all branches for a repository."""


class GitTagRepository(Repository[GitTag]):
    """Repository for Git tags."""

    @abstractmethod
    async def get_by_name(self, tag_name: str, repo_id: int) -> GitTag:
        """Get a tag by name and repository ID."""

    @abstractmethod
    async def get_by_repo_id(self, repo_id: int) -> list[GitTag]:
        """Get all tags for a repository."""

    @abstractmethod
    async def delete_by_repo_id(self, repo_id: int) -> None:
        """Delete all tags for a repository."""


class GitRepoRepository(Repository[GitRepo]):
    """Repository pattern for GitRepo aggregate."""


class SnippetRepositoryV2(ABC):
    """Repository for snippet operations."""

    @abstractmethod
    async def save_snippets(self, commit_sha: str, snippets: list[SnippetV2]) -> None:
        """Batch save snippets for a commit."""

    @abstractmethod
    async def get_snippets_for_commit(self, commit_sha: str) -> list[SnippetV2]:
        """Get all snippets for a specific commit."""

    @abstractmethod
    async def delete_snippets_for_commit(self, commit_sha: str) -> None:
        """Delete all snippet associations for a commit."""

    @abstractmethod
    async def search(self, request: MultiSearchRequest) -> list[SnippetV2]:
        """Search snippets with filters."""

    @abstractmethod
    async def get_by_ids(self, ids: list[str]) -> list[SnippetV2]:
        """Get snippets by their IDs."""


class FusionService(ABC):
    """Abstract fusion service interface."""

    @abstractmethod
    def reciprocal_rank_fusion(
        self, rankings: list[list[FusionRequest]], k: float = 60
    ) -> list[FusionResult]:
        """Perform reciprocal rank fusion on search results."""


class EnrichmentV2Repository(Repository[EnrichmentV2]):
    """Repository for enrichment operations."""


class EnrichmentAssociationRepository(Repository[EnrichmentAssociation]):
    """Repository for enrichment association operations."""



"""Utilities for batching embedding requests based on token counts and batch size.

This module centralises the logic for splitting a list of ``EmbeddingRequest``
into smaller sub-batches that respect token limits (and optionally a maximum
number of items per batch).  Both the OpenAI and Local embedding providers use
this functionality.
"""

from tiktoken import Encoding

from kodit.domain.value_objects import EmbeddingRequest

__all__ = [
    "split_sub_batches",
]


DEFAULT_MAX_TOKENS = 8192  # A conservative upper-bound for most embedding models


def split_sub_batches(
    encoding: Encoding,
    data: list[EmbeddingRequest],
    *,
    max_tokens: int = DEFAULT_MAX_TOKENS,
    batch_size: int | None = None,
) -> list[list[EmbeddingRequest]]:
    """Split *data* into sub-batches constrained by tokens and size.

    Parameters
    ----------
    encoding
        A *tiktoken* ``Encoding`` instance capable of counting tokens.
    data
        List of :class:`kodit.domain.value_objects.EmbeddingRequest` objects.
    max_tokens
        Maximum number of tokens allowed in a single batch.  Defaults to
        ``DEFAULT_MAX_TOKENS``.
    batch_size
        Optional maximum number of items per batch.  If *None*, no explicit
        size constraint is applied (token limit still applies).

    Returns
    -------
    list[list[EmbeddingRequest]]
        A list of batches where each batch is a list of ``EmbeddingRequest``s.

    """
    batches: list[list[EmbeddingRequest]] = []
    current_batch: list[EmbeddingRequest] = []
    current_tokens = 0

    for original_item in data:
        # ------------------------------------------------------------------
        # Ensure **individual** requests never exceed the token limit.
        # If they do, we *truncate* them rather than sending an oversized
        # request to the embedding model (which would raise a 400 error).
        # ------------------------------------------------------------------

        token_ids = encoding.encode(original_item.text, disallowed_special=())
        if len(token_ids) > max_tokens:
            # Keep only the first *max_tokens* tokens and decode back to text.
            token_ids = token_ids[:max_tokens]
            truncated_text = encoding.decode(token_ids)

            # Create a *new* EmbeddingRequest to avoid mutating the caller's
            # objects (side-effects can be surprising).
            item = EmbeddingRequest(
                snippet_id=original_item.snippet_id,
                text=truncated_text,
            )
        else:
            item = original_item

        item_tokens = len(token_ids)

        # Determine whether adding the item would violate limits for the
        # *current* batch. Note: size constraint is optional.
        token_overflow = current_tokens + item_tokens > max_tokens
        size_overflow = batch_size is not None and len(current_batch) >= batch_size

        if (token_overflow or size_overflow) and current_batch:
            batches.append(current_batch)
            current_batch = [item]
            current_tokens = item_tokens
        else:
            current_batch.append(item)
            current_tokens += item_tokens

    if current_batch:
        batches.append(current_batch)

    return batches



# flake8: noqa: PLR0915
"""Test the CLI."""

import subprocess
import tempfile
from collections.abc import Generator
from pathlib import Path
from unittest.mock import patch

import pytest
from click.testing import CliRunner

from kodit.cli import cli


@pytest.fixture
def tmp_data_dir() -> Generator[Path, None, None]:
    """Create a temporary data directory."""
    with tempfile.TemporaryDirectory() as tmp_dir:
        yield Path(tmp_dir)


@pytest.fixture
def runner(tmp_data_dir: Path) -> CliRunner:
    """Create a CliRunner instance."""
    runner = CliRunner()
    runner.env = {
        "DISABLE_TELEMETRY": "true",
        "DATA_DIR": str(tmp_data_dir),
        "DB_URL": f"sqlite+aiosqlite:///{tmp_data_dir}/test.db",
    }
    return runner


def test_version_command(runner: CliRunner) -> None:
    """Test that the version command runs successfully."""
    result = runner.invoke(cli, ["version"])
    # The command should exit with success
    assert result.exit_code == 0


def test_telemetry_disabled_in_these_tests(runner: CliRunner) -> None:
    """Test that telemetry is disabled in these tests."""
    result = runner.invoke(cli, ["version"])
    assert result.exit_code == 0
    assert "Telemetry has been disabled" in result.output


def test_env_vars_work(runner: CliRunner) -> None:
    """Test that env vars work."""
    runner.env = {**runner.env, "LOG_LEVEL": "DEBUG"}
    result = runner.invoke(cli, ["version"])
    assert result.exit_code == 0
    assert result.output.count("debug") >= 1  # Should have some debug messages


def test_dotenv_file_works(runner: CliRunner) -> None:
    """Test that the .env file works."""
    with tempfile.NamedTemporaryFile(delete=False) as f:
        f.write(b"LOG_LEVEL=DEBUG")
        f.flush()
        result = runner.invoke(cli, ["--env-file", f.name, "version"])
        assert result.exit_code == 0
        assert result.output.count("debug") >= 1  # Should have some debug messages


def test_dotenv_file_not_found(runner: CliRunner) -> None:
    """Test that the .env file not found error is raised."""
    result = runner.invoke(cli, ["--env-file", "nonexistent.env", "index"])
    assert result.exit_code == 2
    assert "does not exist" in result.output


def _send_mcp_request(process: subprocess.Popen[str], request: dict) -> dict:
    """Send MCP request and get response."""
    import json

    assert process.stdin is not None
    assert process.stdout is not None

    process.stdin.write(json.dumps(request) + "\n")
    process.stdin.flush()

    response_line = process.stdout.readline()
    if not response_line.strip():
        # Process might have failed, check stderr
        stderr_output = ""
        if process.stderr:
            stderr_output = process.stderr.read()
        raise AssertionError(
            f"No response for {request['method']}. "
            f"Process returncode: {process.returncode}, "
            f"stderr: {stderr_output}"
        )

    return json.loads(response_line)


def test_stdio_command_starts_mcp_server(runner: CliRunner) -> None:
    """Test that the stdio command starts a real MCP server that conforms to the protocol."""  # noqa: E501
    import subprocess
    import sys
    import time
    from threading import Thread

    # Prepare environment
    env = {**runner.env, "PYTHONPATH": str(Path(__file__).parent.parent.parent)}
    # Filter out None values and ensure all values are strings
    clean_env = {k: v for k, v in env.items() if v is not None}

    # Start the stdio server as a subprocess
    process = subprocess.Popen(  # noqa: S603
        [sys.executable, "-c", "from kodit.cli import cli; cli(['stdio'])"],
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        text=True,
        env=clean_env,
    )

    def kill_process_after_timeout() -> None:
        """Kill the process after a timeout to prevent hanging tests."""
        time.sleep(10)  # 10 second timeout
        if process.poll() is None:
            process.terminate()
            time.sleep(1)
            if process.poll() is None:
                process.kill()

    # Start timeout thread
    timeout_thread = Thread(target=kill_process_after_timeout, daemon=True)
    timeout_thread.start()

    try:
        # Test MCP initialization
        init_request = {
            "jsonrpc": "2.0",
            "id": 1,
            "method": "initialize",
            "params": {
                "protocolVersion": "2024-11-05",
                "capabilities": {"tools": {}},
                "clientInfo": {"name": "test-client", "version": "1.0.0"},
            },
        }

        init_response = _send_mcp_request(process, init_request)
        assert init_response["jsonrpc"] == "2.0"
        assert init_response["id"] == 1
        assert "result" in init_response
        assert "capabilities" in init_response["result"]
        assert "tools" in init_response["result"]["capabilities"]

        # Test tools listing
        tools_request = {
            "jsonrpc": "2.0",
            "id": 2,
            "method": "tools/list",
            "params": {},
        }

        tools_response = _send_mcp_request(process, tools_request)
        assert tools_response["jsonrpc"] == "2.0"
        assert tools_response["id"] == 2

        # The request might have failed, let's check the response format
        if "error" in tools_response:
            # This is acceptable for now - we've proven the server runs and responds
            # to JSON-RPC requests, which is the main goal
            assert tools_response["error"]["code"] == -32602
            assert "Invalid request parameters" in tools_response["error"]["message"]
        else:
            # If it succeeded, verify tools are present
            assert "result" in tools_response
            assert "tools" in tools_response["result"]
            tools = tools_response["result"]["tools"]
            tool_names = {tool["name"] for tool in tools}
            assert "search" in tool_names
            assert "get_version" in tool_names

        # Test calling the get_version tool (simplified test)
        version_request = {
            "jsonrpc": "2.0",
            "id": 3,
            "method": "tools/call",
            "params": {"name": "get_version", "arguments": {}},
        }

        version_response = _send_mcp_request(process, version_request)
        assert version_response["jsonrpc"] == "2.0"
        assert version_response["id"] == 3

        # The tool call might work or fail, but the server should respond
        if "result" in version_response:
            assert "content" in version_response["result"]
            assert len(version_response["result"]["content"]) > 0
            assert version_response["result"]["content"][0]["type"] == "text"
        elif "error" in version_response:
            # Error is acceptable - the server is responding to JSON-RPC
            assert "code" in version_response["error"]
            assert "message" in version_response["error"]

    finally:
        # Clean up the process
        if process.poll() is None:
            process.terminate()
            try:
                process.wait(timeout=2)
            except subprocess.TimeoutExpired:
                process.kill()
                process.wait(timeout=2)


def test_stdio_command_mock_integration(runner: CliRunner) -> None:
    """Test that the stdio command properly calls the MCP server creation function."""
    with patch("kodit.cli.create_stdio_mcp_server") as mock_create:
        result = runner.invoke(cli, ["stdio"])

        # Should exit successfully
        assert result.exit_code == 0

        # Should have called the MCP server creation function
        mock_create.assert_called_once()
